import functools
import importlib
import logging
from typing import Dict, Optional

from gaphas.guide import GuidePainter
from gaphas.painter import FreeHandPainter, HandlePainter, PainterChain
from gaphas.segment import LineSegmentPainter
from gaphas.tool.rubberband import RubberbandPainter, RubberbandState
from gaphas.view import GtkView
from gi.repository import Gdk, GdkPixbuf, Gtk

from gaphor.core import event_handler, gettext
from gaphor.core.modeling import StyleSheet
from gaphor.core.modeling.diagram import Diagram, StyledDiagram
from gaphor.core.modeling.event import AttributeUpdated, ElementDeleted
from gaphor.diagram.diagramtoolbox import get_tool_def, tooliter
from gaphor.diagram.drop import drop
from gaphor.diagram.painter import DiagramTypePainter, ItemPainter
from gaphor.diagram.selection import Selection
from gaphor.diagram.tools import (
    apply_default_tool_set,
    apply_magnet_tool_set,
    apply_placement_tool_set,
)
from gaphor.diagram.tools.magnet import MagnetPainter
from gaphor.diagram.tools.placement import create_item, open_editor
from gaphor.event import Notification
from gaphor.transaction import Transaction
from gaphor.ui.event import DiagramClosed, DiagramSelectionChanged, ToolSelected

log = logging.getLogger(__name__)


@functools.lru_cache(maxsize=1)
def placement_icon_base():
    f = importlib.resources.files("gaphor") / "ui" / "placement-icon-base.png"
    return GdkPixbuf.Pixbuf.new_from_file_at_scale(str(f), 64, 64, True)


if hasattr(GtkView, "set_css_name"):
    GtkView.set_css_name("diagramview")


if Gtk.get_major_version() == 3:
    _placement_pixbuf_map: Dict[str, GdkPixbuf.Pixbuf] = {}

    def get_placement_cursor(display, icon_name):
        if icon_name in _placement_pixbuf_map:
            pixbuf = _placement_pixbuf_map[icon_name]
        else:
            pixbuf = placement_icon_base().copy()
            icon = Gtk.IconTheme.get_default().load_icon(icon_name, 24, 0)
            icon.copy_area(
                0,
                0,
                icon.get_width(),
                icon.get_height(),
                pixbuf,
                9,
                15,
            )
            _placement_pixbuf_map[icon_name] = pixbuf
        return Gdk.Cursor.new_from_pixbuf(display, pixbuf, 1, 1)

else:

    @functools.lru_cache(maxsize=None)
    def get_placement_icon(display, icon_name):
        if display is None:
            display = Gdk.Display.get_default()
        pixbuf = placement_icon_base().copy()
        theme_icon = Gtk.IconTheme.get_for_display(display).lookup_icon(
            icon_name,
            None,
            24,
            1,
            Gtk.TextDirection.NONE,
            Gtk.IconLookupFlags.FORCE_SYMBOLIC,
        )
        icon = GdkPixbuf.Pixbuf.new_from_file_at_scale(
            theme_icon.get_file().get_path(), 32, 32, True
        )
        icon.copy_area(
            0,
            0,
            icon.get_width(),
            icon.get_height(),
            pixbuf,
            9,
            15,
        )
        return Gdk.Texture.new_for_pixbuf(pixbuf)

    def get_placement_cursor(display, icon_name):
        return Gdk.Cursor.new_from_texture(get_placement_icon(display, icon_name), 1, 1)


class DiagramPage:

    if Gtk.get_major_version() == 3:
        VIEW_TARGET_STRING = 0
        VIEW_TARGET_ELEMENT_ID = 1
        VIEW_TARGET_TOOLBOX_ACTION = 2
        VIEW_DND_TARGETS = [
            Gtk.TargetEntry.new("gaphor/element-id", 0, VIEW_TARGET_ELEMENT_ID),
            Gtk.TargetEntry.new("gaphor/toolbox-action", 0, VIEW_TARGET_TOOLBOX_ACTION),
        ]

    def __init__(
        self, diagram, event_manager, element_factory, properties, modeling_language
    ):
        self.event_manager = event_manager
        self.element_factory = element_factory
        self.properties = properties
        self.diagram = diagram
        self.modeling_language = modeling_language

        self.view: Optional[GtkView] = None
        self.widget: Optional[Gtk.Widget] = None
        self.diagram_css: Optional[Gtk.CssProvider] = None

        self.rubberband_state = RubberbandState()

        self.event_manager.subscribe(self._on_element_delete)
        self.event_manager.subscribe(self._on_attribute_updated)
        self.event_manager.subscribe(self._on_tool_selected)

    title = property(lambda s: s.diagram and s.diagram.name or gettext("<None>"))

    def get_diagram(self):
        return self.diagram

    def get_view(self):
        return self.view

    def construct(self):
        """Create the widget.

        Returns: the newly created widget.
        """
        assert self.diagram

        view = GtkView(selection=Selection())
        if Gtk.get_major_version() == 3:
            view.add_events(Gdk.EventMask.SMOOTH_SCROLL_MASK)
            view.drag_dest_set(
                Gtk.DestDefaults.ALL,
                DiagramPage.VIEW_DND_TARGETS,
                Gdk.DragAction.MOVE | Gdk.DragAction.COPY | Gdk.DragAction.LINK,
            )

        self.diagram_css = Gtk.CssProvider.new()
        view.get_style_context().add_provider(
            self.diagram_css, Gtk.STYLE_PROVIDER_PRIORITY_USER
        )

        scrolled_window = Gtk.ScrolledWindow()
        scrolled_window.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        if Gtk.get_major_version() == 3:
            scrolled_window.add(view)
            scrolled_window.show_all()
            view.connect("drag-data-received", self._on_drag_data_received)
        else:
            scrolled_window.set_child(view)

        view.selection.add_handler(self._on_view_selection_changed)

        self.view = view
        self.widget = scrolled_window

        self.select_tool("toolbox-pointer")

        self.set_drawing_style()

        # Set model only after the painters are set
        view.model = self.diagram

        return self.widget

    def apply_tool_set(self, tool_name):
        """Return a tool associated with an id (action name)."""
        if tool_name == "toolbox-pointer":
            return apply_default_tool_set(
                self.view,
                self.modeling_language,
                self.event_manager,
                self.rubberband_state,
            )
        elif tool_name == "toolbox-magnet":
            return apply_magnet_tool_set(
                self.view,
                self.modeling_language,
                self.event_manager,
            )

        tool_def = get_tool_def(self.modeling_language, tool_name)
        item_factory = tool_def.item_factory
        handle_index = tool_def.handle_index
        return apply_placement_tool_set(
            self.view,
            item_factory=item_factory,
            modeling_language=self.modeling_language,
            event_manager=self.event_manager,
            handle_index=handle_index,
        )

    def get_tool_icon_name(self, tool_name):
        if tool_name in ("toolbox-pointer", "toolbox-magnet"):
            return None
        return next(
            t
            for t in tooliter(self.modeling_language.toolbox_definition)
            if t.id == tool_name
        ).icon_name

    @event_handler(ToolSelected)
    def _on_tool_selected(self, event: ToolSelected):
        self.select_tool(event.tool_name)

    @event_handler(ElementDeleted)
    def _on_element_delete(self, event: ElementDeleted):
        if event.element is self.diagram:
            self.event_manager.handle(DiagramClosed(self.diagram))

    @event_handler(AttributeUpdated)
    def _on_attribute_updated(self, event: AttributeUpdated):
        if event.property is StyleSheet.styleSheet:
            self.set_drawing_style()

            diagram = self.diagram
            for item in diagram.get_all_items():
                diagram.request_update(item)
        elif event.property is Diagram.name and self.view:
            self.view.update_back_buffer()

    def close(self):
        """Tab is destroyed.

        Do the same thing that would be done if Close was pressed.
        """
        assert self.widget
        if Gtk.get_major_version() == 3:
            self.widget.destroy()

        self.event_manager.unsubscribe(self._on_element_delete)
        self.event_manager.unsubscribe(self._on_attribute_updated)
        self.event_manager.unsubscribe(self._on_tool_selected)
        self.view = None

    def select_tool(self, tool_name: str):
        if not self.view:
            return
        self.apply_tool_set(tool_name)
        icon_name = self.get_tool_icon_name(tool_name)
        if Gtk.get_major_version() == 3:
            window = self.view.get_window()
            if icon_name and window:
                window.set_cursor(get_placement_cursor(window.get_display(), icon_name))
            elif window:
                window.set_cursor(None)
        elif icon_name:
            self.view.set_cursor(get_placement_cursor(None, icon_name))
        else:
            self.view.set_cursor(None)

    def set_drawing_style(self):
        """Set the drawing style for the diagram based on the active style
        sheet."""
        assert self.view
        assert self.diagram_css

        style = self.diagram.style(StyledDiagram(self.diagram))

        bg = style.get("background-color")
        # Default background to white, slightly gray in dark mode
        if not bg or bg[3] == 0.0:
            bg = (1.0, 1.0, 1.0, 0.94)
        self.diagram_css.load_from_data(
            f"diagramview {{ background-color: rgba({int(255*bg[0])}, {int(255*bg[1])}, {int(255*bg[2])}, {bg[3]}); }}".encode()
        )

        view = self.view

        item_painter = ItemPainter(view.selection)

        if sloppiness := style.get("line-style", 0.0):
            item_painter = FreeHandPainter(item_painter, sloppiness=sloppiness)

        view.bounding_box_painter = item_painter
        view.painter = (
            PainterChain()
            .append(item_painter)
            .append(HandlePainter(view))
            .append(LineSegmentPainter(view.selection))
            .append(GuidePainter(view))
            .append(MagnetPainter(view))
            .append(RubberbandPainter(self.rubberband_state))
            .append(DiagramTypePainter(self.diagram))
        )

        view.request_update(self.diagram.get_all_items())

    def _on_view_selection_changed(self, item):
        view = self.view
        if not view:
            return
        selection = view.selection
        self.event_manager.handle(
            DiagramSelectionChanged(
                view, selection.focused_item, selection.selected_items
            )
        )

    if Gtk.get_major_version() == 3:

        def _on_drag_data_received(self, view, context, x, y, data, info, time):
            """Handle data dropped on the diagram."""
            if (
                data
                and data.get_format() == 8
                and info == DiagramPage.VIEW_TARGET_TOOLBOX_ACTION
            ):
                tool_def = get_tool_def(
                    self.modeling_language, data.get_data().decode()
                )
                with Transaction(self.event_manager):
                    item = create_item(view, tool_def.item_factory, x, y)
                open_editor(item, view, self.event_manager)
                context.finish(True, False, time)
            elif (
                data
                and data.get_format() == 8
                and info == DiagramPage.VIEW_TARGET_ELEMENT_ID
            ):
                element_id = data.get_data().decode()
                element = self.element_factory.lookup(element_id)
                assert element

                x, y = view.matrix.inverse().transform_point(x, y)
                with Transaction(self.event_manager):
                    if item := drop(element, self.diagram, x, y):
                        view.selection.unselect_all()
                        view.selection.focused_item = item

                        context.finish(True, False, time)
                    else:
                        self.event_manager.handle(
                            Notification(
                                gettext("Element can’t be represented on a diagram.")
                            )
                        )
                        context.finish(False, False, time)
            else:
                context.finish(False, False, time)



from __future__ import annotations

from functools import singledispatch
from typing import Iterable, Iterator

import pydot
from gaphas.connector import ConnectionSink, Connector
from gaphas.geometry import Point, Rect
from gaphas.item import NW
from gaphas.matrix import Matrix
from gaphas.segment import Segment

from gaphor.abc import ActionProvider, Service
from gaphor.action import action
from gaphor.core.modeling import Diagram, Element, Presentation
from gaphor.diagram.connectors import ItemTemporaryDisconnected
from gaphor.diagram.presentation import (
    AttachedPresentation,
    ElementPresentation,
    HandlePositionEvent,
    LinePresentation,
)
from gaphor.i18n import gettext
from gaphor.transaction import Transaction
from gaphor.UML import NamedElement
from gaphor.UML.actions.activitynodes import ForkNodeItem

DOT = "dot"
DPI = 72.0


class AutoLayoutService(Service, ActionProvider):
    def __init__(self, event_manager, diagrams, tools_menu=None, dump_gv=False):
        self.event_manager = event_manager
        self.diagrams = diagrams
        if tools_menu:
            tools_menu.add_actions(self)
        self.dump_gv = dump_gv

    def shutdown(self):
        pass

    @action(
        name="auto-layout", label=gettext("Auto Layout"), shortcut="<Primary><Shift>L"
    )
    def layout_current_diagram(self):
        if current_diagram := self.diagrams.get_current_diagram():
            self.layout(current_diagram)

    @action(
        name="auto-layout-ortho",
        label=gettext("Auto Layout (orthogonal)"),
        shortcut="<Primary><Shift>K",
    )
    def layout_current_diagram_orthogonal(self):
        if current_diagram := self.diagrams.get_current_diagram():
            self.layout(current_diagram, splines="ortho")

    def layout(self, diagram: Diagram, splines="polyline"):
        auto_layout = AutoLayout(self.event_manager, self.dump_gv)

        with Transaction(self.event_manager):
            auto_layout.layout(diagram, splines)


class AutoLayout:
    def __init__(self, event_manager=None, dump_gv=False) -> None:
        self.event_manager = event_manager
        self.dump_gv = dump_gv

    def layout(self, diagram: Diagram, splines="polyline") -> None:
        diagram.update_now(diagram.get_all_items())
        graph = diagram_as_pydot(diagram, splines=splines)
        rendered_graph = self.render(graph)
        self.apply_layout(diagram, rendered_graph)

    def render(self, graph: pydot.Dot):
        if self.dump_gv:
            graph.write("auto_layout.gv")

        rendered_string = graph.create(prog=DOT, format="dot", encoding="utf-8").decode(
            "utf-8"
        )

        rendered_graphs = pydot.graph_from_dot_data(rendered_string)
        return rendered_graphs[0]

    def apply_layout(  # noqa: C901
        self, diagram, rendered_graph, parent_presentation=None, height=None
    ):
        if height is None:
            _, _, _, height = parse_bb(rendered_graph.get_node("graph")[0].get("bb"))

        matrix_c2i = (
            parent_presentation.matrix_i2c.inverse()
            if parent_presentation
            else Matrix()
        )

        # First record original positions for involved lines
        for edge in rendered_graph.get_edges():
            if presentation := presentation_for_object(diagram, edge):
                for handle in (presentation.head, presentation.tail):
                    if cinfo := diagram.connections.get_connection(handle):
                        self.handle(
                            ItemTemporaryDisconnected(
                                presentation, handle, cinfo.connected, cinfo.port
                            )
                        )

                for handle in presentation.handles():
                    self.handle(
                        HandlePositionEvent(presentation, handle, handle.pos.tuple())
                    )

        for subgraph in rendered_graph.get_subgraphs():
            if presentation := presentation_for_object(
                diagram, subgraph.get_node("graph")[0]
            ):
                if bb := subgraph.get_node("graph")[0].get("bb"):
                    x, y, w, h = parse_bb(bb, height)
                    presentation.handles()[NW].pos = (0.0, 0.0)
                    presentation.width = w
                    presentation.height = h

                    new_pos = matrix_c2i.transform_point(x, y)
                    presentation.matrix.set(
                        x0=new_pos[0],
                        y0=new_pos[1],
                    )
                    self.apply_layout(
                        diagram,
                        subgraph,
                        parent_presentation=presentation,
                        height=height,
                    )

        for node in rendered_graph.get_nodes():
            if not node.get_pos():
                continue

            if presentation := presentation_for_object(diagram, node):
                center = parse_point(node.get_pos(), height)
                if isinstance(presentation, ElementPresentation):
                    # Normalize handle placement
                    w = presentation.width
                    h = presentation.height
                    presentation.handles()[NW].pos = (0.0, 0.0)
                    presentation.width = w
                    presentation.height = h

                    new_pos = matrix_c2i.transform_point(
                        center[0] - w / 2, center[1] - h / 2
                    )
                else:
                    new_pos = matrix_c2i.transform_point(center[0], center[1])
                presentation.matrix.set(
                    x0=new_pos[0],
                    y0=new_pos[1],
                )
                if isinstance(presentation, AttachedPresentation):
                    reconnect(
                        presentation, presentation.handles()[0], diagram.connections
                    )

        for edge in rendered_graph.get_edges():
            if presentation := presentation_for_object(diagram, edge):
                presentation.orthogonal = False

                points = parse_edge_pos(edge.get_pos(), height)
                segment = Segment(presentation, diagram)
                while len(points) > len(presentation.handles()):
                    segment.split_segment(0)
                while len(points) < len(presentation.handles()):
                    segment.merge_segment(0)

                assert len(points) == len(presentation.handles())

                matrix = presentation.matrix_i2c.inverse()
                for handle, point in zip(presentation.handles(), points):
                    handle.pos = matrix.transform_point(*point)

                for handle in (presentation.head, presentation.tail):
                    reconnect(presentation, handle, diagram.connections)

    def handle(self, event):
        if self.event_manager:
            self.event_manager.handle(event)


def presentation_for_object(diagram, obj) -> Presentation | None:
    if not obj.get("id"):
        return None

    id = strip_quotes(obj.get("id"))
    return next((p for p in diagram.ownedPresentation if p.id == id), None)


def reconnect(presentation, handle, connections) -> None:
    if not (connected := connections.get_connection(handle)):
        return

    connector = Connector(presentation, handle, connections)
    sink = ConnectionSink(connected.connected, distance=float("inf"))
    connector.connect(sink)


def diagram_as_pydot(diagram: Diagram, splines: str) -> pydot.Dot:
    graph = pydot.Dot(
        "gaphor", graph_type="digraph", compound="true", pad=8 / DPI, splines=splines
    )

    for presentation in diagram.ownedPresentation:
        if presentation.parent:
            continue

        add_to_graph(graph, as_pydot(presentation))

    return graph


@singledispatch
def as_pydot(element: Element) -> Iterator[pydot.Common]:
    return iter(())


@as_pydot.register
def _(presentation: ElementPresentation):
    if as_cluster(presentation):
        graph = pydot.Cluster(
            presentation.id,
            id=presentation.id,
            label=f"{presentation.subject.name}\n\n\n"
            if isinstance(presentation.subject, NamedElement)
            else "\n\n",
            margin=20,
        )

        # Add a placeholder, so we can connect to the cluster
        graph.add_node(
            pydot.Node(
                f'"{presentation.id}"',
                label="",
                shape="point",
            )
        )

        for child in presentation.children:
            if isinstance(child, AttachedPresentation):
                yield as_pydot(child)
            else:
                add_to_graph(graph, as_pydot(child))

        yield graph
    else:
        for attached in presentation.children:
            if isinstance(attached, AttachedPresentation):
                yield as_pydot(attached)

        yield pydot.Node(
            f'"{presentation.id}"',
            id=presentation.id,
            label="",
            shape="rect",
            width=presentation.width / DPI,
            height=presentation.height / DPI,
        )


@as_pydot.register
def _(presentation: LinePresentation):
    connections = presentation.diagram.connections
    head_connection = connections.get_connection(presentation.head)
    tail_connection = connections.get_connection(presentation.tail)
    if (
        head_connection
        and next(as_pydot(head_connection.connected), None)
        and tail_connection
        and next(as_pydot(tail_connection.connected), None)
    ):
        extra_args = {}
        if as_cluster(head_connection.connected):
            extra_args["lhead"] = f"cluster_{head_connection.connected.id}"
        if as_cluster(tail_connection.connected):
            extra_args["ltail"] = f"cluster_{tail_connection.connected.id}"

        yield pydot.Edge(
            tail_connection.connected.id,
            head_connection.connected.id,
            id=presentation.id,
            minlen=3,
            arrowhead="none",
            **extra_args,
        )


@as_pydot.register
def _(presentation: AttachedPresentation):
    yield pydot.Node(
        presentation.id,
        id=presentation.id,
        label="",
        shape="point",
    )
    handle = presentation.handles()[0]
    if connection := presentation.diagram.connections.get_connection(handle):
        yield pydot.Edge(
            connection.connected.id,
            presentation.id,
            len=0.01,
        )


@as_pydot.register
def _(presentation: ForkNodeItem):
    h1, h2 = presentation.handles()
    yield pydot.Node(
        presentation.id,
        id=presentation.id,
        label="",
        shape="rect",
        width=(h2.pos.x - h1.pos.x) / DPI,
        height=(h2.pos.y - h1.pos.y) / DPI,
    )


def as_cluster(presentation: Presentation):
    return presentation.children and not all(
        isinstance(c, AttachedPresentation) for c in presentation.children
    )


def add_to_graph(graph, edge_or_node) -> None:
    if isinstance(edge_or_node, pydot.Edge):
        graph.add_edge(edge_or_node)
    elif isinstance(edge_or_node, pydot.Node):
        graph.add_node(edge_or_node)
    elif isinstance(edge_or_node, pydot.Graph):
        graph.add_subgraph(edge_or_node)
    elif isinstance(edge_or_node, Iterable):
        for obj in edge_or_node:
            add_to_graph(graph, obj)
    elif edge_or_node:
        raise ValueError(f"Can't transform {edge_or_node} to something DOT'ish?")


def parse_edge_pos(pos_str: str, height: float) -> list[Point]:
    raw_points = strip_quotes(pos_str).split(" ")

    points = [parse_point(raw_points.pop(0), height)]

    while raw_points:
        # Drop bezier curve support points
        raw_points.pop(0)
        raw_points.pop(0)
        points.append(parse_point(raw_points.pop(0), height))
    points.reverse()
    return points


def parse_point(point, height) -> Point:
    x, y = strip_quotes(point).split(",")
    return (float(x), height - float(y))


def parse_bb(bb, height: float | None = None) -> Rect:
    llx, lly, urx, ury = map(float, strip_quotes(bb).split(","))
    return (llx, lly if height is None else height - ury, urx - llx, ury - lly)


def strip_quotes(s):
    return s.replace('"', "").replace("\\\n", "")



import xml.sax.handler
from typing import Dict, List, Tuple
from xml.sax.saxutils import escape, quoteattr

# See whether the xmlcharrefreplace error handler is
# supported
try:
    from codecs import xmlcharrefreplace_errors

    _error_handling = "xmlcharrefreplace"
    del xmlcharrefreplace_errors
except ImportError:
    _error_handling = "strict"


class XMLWriter(xml.sax.handler.ContentHandler):
    def __init__(self, out, encoding="utf-8"):
        super().__init__()
        self._out = out
        self._encoding = encoding
        self._ns_contexts: List[Dict[str, str]] = [{}]  # contains uri -> prefix dicts
        self._current_context = self._ns_contexts[-1]
        self._undeclared_ns_maps: List[Tuple[str, str]] = []

        self._in_cdata = False
        self._in_start_tag = False
        self._next_newline = False

    def _write(self, text, start_tag=False, end_tag=False):
        """Write data.

        Tags should not be escaped. They should be marked by setting
        either ``start_tag`` or ``end_tag`` to ``True``. Only the tag
        should be marked this way. Other stuff, such as namespaces and
        attributes can be written directly to the file.
        """
        if not isinstance(text, str):
            text = text.decode(self._encoding, _error_handling)

        if self._next_newline:
            self._out.write("\n")
            self._next_newline = False

        if start_tag and not self._in_start_tag:
            self._in_start_tag = True
            self._out.write("<")
        elif start_tag:
            self._out.write(">")
            self._out.write("\n")
            self._out.write("<")
        elif end_tag and self._in_start_tag:
            self._out.write("/>")
            self._in_start_tag = False
            self._next_newline = True
            return
        elif self._in_start_tag:
            self._out.write(">")
            self._in_start_tag = False
        elif end_tag:
            self._out.write("</")
            self._out.write(text)
            self._out.write(">")
            self._in_start_tag = False
            self._next_newline = True
            return

        self._out.write(text)

    def _qname(self, name):
        """Builds a qualified name from a (ns_url, localname) pair."""
        if name[0]:
            if prefix := self._current_context[name[0]]:
                # If it is not the default namespace, prepend the prefix
                return f"{prefix}:{name[1]}"
        # Return the unqualified name
        return name[1]

    # ContentHandler methods

    def startDocument(self):
        self._write(f'<?xml version="1.0" encoding="{self._encoding}"?>\n')

    def startPrefixMapping(self, prefix, uri):
        self._ns_contexts.append(self._current_context.copy())
        self._current_context[uri] = prefix
        self._undeclared_ns_maps.append((prefix, uri))

    def endPrefixMapping(self, prefix):
        self._current_context = self._ns_contexts[-1]
        del self._ns_contexts[-1]

    def startElement(self, name, attrs):
        self._write(name, start_tag=True)
        for (name, value) in list(attrs.items()):
            self._out.write(f" {name}={quoteattr(value)}")

    def endElement(self, name):
        self._write(name, end_tag=True)

    def startElementNS(self, name, qname, attrs):
        self._write(self._qname(name), start_tag=True)

        for prefix, uri in self._undeclared_ns_maps:
            if prefix:
                self._out.write(f' xmlns:{prefix}="{uri}"')
            else:
                self._out.write(f' xmlns="{uri}"')
        self._undeclared_ns_maps = []

        for (name, value) in list(attrs.items()):
            self._out.write(f" {self._qname(name)}={quoteattr(value)}")

    def endElementNS(self, name, qname):
        self._write(f"{self._qname(name)}", end_tag=True)

    def characters(self, content):
        if self._in_cdata:
            self._write(content.replace("]]>", "] ]>"))
        else:
            self._write(escape(content))

    def ignorableWhitespace(self, content):
        self._write(content)

    def processingInstruction(self, target, data):
        self._write(f"<?{target} {data}?>")

    def comment(self, comment):
        self._write("<!-- ")
        self._write(comment.replace("-->", "- ->"))
        self._write(" -->")

    def startCDATA(self):
        self._write("<![CDATA[")
        self._in_cdata = True

    def endCDATA(self):
        self._write("]]>")
        self._in_cdata = False



"""This module contains user interface related code, such as the main screen
and diagram windows."""

import logging
import os
import sys
from typing import Optional

import darkdetect
import gi

if os.getenv("GAPHOR_USE_GTK") != "NONE":
    # Allow to explicitly *not* initialize GTK (for docs, mainly)
    gtk_version = "3.0" if os.getenv("GAPHOR_USE_GTK") == "3" else "4.0"
    gtk_source_version = "4" if os.getenv("GAPHOR_USE_GTK") == "3" else "5"

    gi.require_version("Gtk", gtk_version)
    gi.require_version("Gdk", gtk_version)
    gi.require_version("GtkSource", gtk_source_version)
    if gtk_version == "4.0":
        gi.require_version("Adw", "1")
        from gi.repository import Adw


from gi.repository import Gdk, Gio, GLib, Gtk

from gaphor.application import Application, Session, distribution
from gaphor.core import event_handler
from gaphor.event import ActiveSessionChanged, ApplicationShutdown, SessionCreated
from gaphor.ui.actiongroup import apply_application_actions
from gaphor.ui.macosshim import macos_init

APPLICATION_ID = "org.gaphor.Gaphor"
LOG_FORMAT = "%(name)s %(levelname)s %(message)s"


def main(argv=sys.argv) -> int:
    """Start Gaphor from the command line.  This function creates an option
    parser for retrieving arguments and options from the command line.  This
    includes a Gaphor model to load.

    The application is then initialized, passing along the option
    parser.  This provides plugins and services with access to the
    command line options and may add their own.
    """

    def has_option(*options):
        return any(o in argv for o in options)

    if has_option("-v", "--version"):
        print(f"Gaphor {distribution().version}")
        return 0

    if has_option("-d", "--debug"):
        logging.basicConfig(level=logging.INFO, format=LOG_FORMAT)
        logging.getLogger("gaphor").setLevel(logging.DEBUG)
    elif has_option("-q", "--quiet"):
        logging.basicConfig(level=logging.WARNING, format=LOG_FORMAT)
    else:
        logging.basicConfig(level=logging.INFO, format=LOG_FORMAT)

    # Set dark mode for non-FreeDesktop platforms only:
    if sys.platform in ("darwin", "win32"):
        if Gtk.get_major_version() == 3:
            Gtk.Settings.get_default().set_property(
                "gtk-application-prefer-dark-theme", darkdetect.isDark()
            )
        else:
            if darkdetect.isDark():
                color_scheme = Adw.ColorScheme.PREFER_DARK
            else:
                color_scheme = Adw.ColorScheme.PREFER_LIGHT
            Adw.StyleManager.get_default().set_color_scheme(color_scheme)

    if has_option("-p", "--profiler"):

        import cProfile
        import pstats

        with cProfile.Profile() as profile:
            exit_code = profile.runcall(run, argv)

        profile_stats = pstats.Stats(profile)
        profile_stats.strip_dirs().sort_stats("time").print_stats(50)
        return exit_code

    return run(argv)


def run(args: list[str]) -> int:
    application: Optional[Application] = None

    def app_startup(gtk_app):
        nonlocal application

        @event_handler(SessionCreated)
        def on_session_created(event):
            event_manager = event.session.get_service("event_manager")
            event_manager.subscribe(on_session_changed)
            main_window = event.session.get_service("main_window")
            main_window.open(gtk_app)

        @event_handler(ActiveSessionChanged)
        def on_session_changed(event):
            if isinstance(event.service, Session):
                main_window = event.service.get_service("main_window")
                if main_window.window:
                    main_window.window.present_with_time(Gdk.CURRENT_TIME)

        @event_handler(ApplicationShutdown)
        def on_quit(event):
            gtk_app.quit()

        try:
            application = Application()
            apply_application_actions(application, gtk_app)
            macos_init(application)
            event_manager = application.get_service("event_manager")
            event_manager.subscribe(on_session_created)
            event_manager.subscribe(on_quit)
            application.get_service(
                "self_test" if "--self-test" in args else "greeter"
            ).init(gtk_app)
        except Exception:
            gtk_app.exit_code = 1
            gtk_app.quit()
            raise

    def app_activate(gtk_app):
        assert application
        if not application.has_sessions():
            application.get_service("greeter").open()

    def app_open(gtk_app, files, n_files, hint):
        # appfilemanager should take care of this:
        assert application
        if hint == "__new__":
            application.new_session()
        else:
            for file in files:
                application.new_session(filename=file.get_path())

    if gtk_version == "3.0":
        gtk_app = Gtk.Application(
            application_id=APPLICATION_ID, flags=Gio.ApplicationFlags.HANDLES_OPEN
        )
    else:
        gtk_app = Adw.Application(
            application_id=APPLICATION_ID, flags=Gio.ApplicationFlags.HANDLES_OPEN
        )
    gtk_app.exit_code = 0
    add_main_options(gtk_app)
    gtk_app.connect("startup", app_startup)
    gtk_app.connect("activate", app_activate)
    gtk_app.connect("open", app_open)
    gtk_app.run(args)
    return gtk_app.exit_code


def add_main_options(gtk_app):
    """These parameters are handled in `gaphor.ui.main()`.

    Define them here, so they show up on `gaphor --help`.
    """
    gtk_app.add_main_option(
        "version",
        ord("v"),
        GLib.OptionFlags.NONE,
        GLib.OptionArg.NONE,
        "Print version and exit",
        None,
    )
    gtk_app.add_main_option(
        "debug",
        ord("d"),
        GLib.OptionFlags.NONE,
        GLib.OptionArg.NONE,
        "Debug output",
        None,
    )
    gtk_app.add_main_option(
        "quiet",
        ord("q"),
        GLib.OptionFlags.NONE,
        GLib.OptionArg.NONE,
        "Quiet output",
        None,
    )
    gtk_app.add_main_option(
        "profiler",
        ord("p"),
        GLib.OptionFlags.NONE,
        GLib.OptionArg.NONE,
        "Run in profiler",
        None,
    )
    gtk_app.add_main_option(
        "self-test",
        0,
        GLib.OptionFlags.NONE,
        GLib.OptionArg.NONE,
        "Run self test and exit",
        None,
    )



try:
    import gi
    from gi.repository import Gtk

    if Gtk.get_major_version() == 3:
        gi.require_version("GtkosxApplication", "1.0")
    else:
        raise ValueError()
except ValueError:

    def macos_init(application):
        pass

else:
    from gi.repository import GtkosxApplication

    macos_app = GtkosxApplication.Application.get()

    def open_file(macos_app, path, application):
        if path == __file__:
            return False

        application.new_session(filename=path)

        return True

    def block_termination(macos_app, application):
        quit = application.quit()
        return not quit

    def macos_init(application):
        macos_app.connect("NSApplicationOpenFile", open_file, application)
        macos_app.connect(
            "NSApplicationBlockTermination", block_termination, application
        )



from __future__ import annotations

import logging

from gi.repository import Gdk, Gtk, Pango

from gaphor import UML
from gaphor.core import gettext, transactional
from gaphor.core.format import format, parse
from gaphor.core.modeling import Diagram
from gaphor.diagram.iconname import get_icon_name
from gaphor.ui.namespacemodel import RELATIONSHIPS

log = logging.getLogger(__name__)


def namespace_view(model: Gtk.TreeModel):
    if Gtk.get_major_version() == 3:
        DND_TARGETS = [
            Gtk.TargetEntry.new(
                "GTK_TREE_MODEL_ROW",
                Gtk.TargetFlags.SAME_APP,
                0,
            ),
            Gtk.TargetEntry.new("gaphor/element-id", Gtk.TargetFlags.SAME_APP, 1),
        ]
    else:
        DND_TARGETS = Gdk.ContentFormats.new(
            ["gaphor/element-id", "GTK_TREE_MODEL_ROW"]
        )

    view = Gtk.TreeView.new_with_model(model)

    view.set_property("headers-visible", False)
    view.set_property("search-column", 0)

    selection = view.get_selection()
    selection.set_mode(Gtk.SelectionMode.BROWSE)
    column = Gtk.TreeViewColumn.new()
    column.set_spacing(2)
    # First cell in the column is for an image...
    cell = Gtk.CellRendererPixbuf()
    column.pack_start(cell, 0)
    column.set_cell_data_func(cell, _set_pixbuf, None)

    # Second cell if for the name of the object...
    cell = Gtk.CellRendererText()
    cell.connect("edited", _text_edited, model)
    column.pack_start(cell, 0)
    column.set_cell_data_func(cell, _set_text, None)

    view.append_column(column)

    if Gtk.get_major_version() == 3:
        view.enable_model_drag_source(
            Gdk.ModifierType.BUTTON1_MASK | Gdk.ModifierType.BUTTON3_MASK,
            DND_TARGETS,
            Gdk.DragAction.COPY | Gdk.DragAction.MOVE,
        )
        view.enable_model_drag_dest(
            DND_TARGETS,
            Gdk.DragAction.MOVE,
        )
        view._controller = tree_view_expand_collapse(view)
    else:
        view.enable_model_drag_source(
            Gdk.ModifierType.BUTTON1_MASK | Gdk.ModifierType.BUTTON3_MASK,
            DND_TARGETS,
            Gdk.DragAction.COPY | Gdk.DragAction.MOVE,
        )
        view.enable_model_drag_dest(
            DND_TARGETS,
            Gdk.DragAction.MOVE,
        )
        tree_view_expand_collapse(view)

    def search_func(model, column, key, rowiter):
        # Note that this function returns `False` for a match!
        assert column == 0
        row = model[rowiter]
        not_matched = True

        # Search in child rows.  If any element in the underlying
        # tree matches, it will expand.
        for inner in row.iterchildren():
            if not search_func(model, column, key, inner.iter):
                view.expand_to_path(row.path)
                not_matched = False

        element = list(row)[column]
        if element is not RELATIONSHIPS:
            s = format(element)
            if s and key.lower() in s.lower():
                not_matched = False

        if not_matched:
            view.collapse_row(row.path)

        return not_matched

    view.set_search_equal_func(search_func)

    return view


def _set_pixbuf(column, cell, model, iter, data):
    element = model.get_value(iter, 0)

    if element is RELATIONSHIPS:
        cell.set_property("icon-name", None)
        cell.set_property("visible", False)
    elif isinstance(element, (UML.Parameter, UML.Property, UML.Operation)):
        cell.set_property("icon-name", None)
        cell.set_property("visible", False)
    else:
        icon_name = get_icon_name(element)
        cell.set_property("visible", True)

        if icon_name:
            cell.set_property("icon-name", icon_name)


def _set_text(column, cell, model, iter, data):
    """Set font and of model elements in tree view."""
    element = model.get_value(iter, 0)

    cell.set_property(
        "weight",
        Pango.Weight.BOLD if isinstance(element, Diagram) else Pango.Weight.NORMAL,
    )

    cell.set_property(
        "style",
        Pango.Style.ITALIC
        if isinstance(element, (UML.Classifier, UML.BehavioralFeature))
        and element.isAbstract
        else Pango.Style.NORMAL,
    )

    if element is RELATIONSHIPS:
        text = gettext("<Relationships>")
    else:
        text = format(element) or gettext("<None>")
    cell.set_property("text", text)


@transactional
def _text_edited(cell, path_str, new_text, model):
    """The text has been edited.

    This method updates the data object. Note that 'path_str' is a
    string where the fields are separated by colons ':', like this:
    '0:1:1'. We first turn them into a tuple.
    """
    iter = model.get_iter_from_string(path_str)
    element = model.get_value(iter, 0)
    try:
        parse(element, new_text)
    except TypeError:
        log.debug(f"No parser for {element}")


def tree_view_expand_collapse(view):
    def on_key_pressed(controller, keyval, keycode, state):
        if keyval == Gdk.KEY_Right:
            path, _column = view.get_cursor()
            view.expand_row(path, state & Gdk.ModifierType.SHIFT_MASK)
            return True
        elif keyval == Gdk.KEY_Left:
            path, _column = view.get_cursor()
            view.collapse_row(path)
            return True

    if Gtk.get_major_version() == 3:
        controller = Gtk.EventControllerKey.new(view)
    else:
        controller = Gtk.EventControllerKey.new()
        view.add_controller(controller)
    controller.connect("key-pressed", on_key_pressed)
    return controller



from __future__ import annotations

import textwrap

from gaphor.core.modeling.element import Element
from gaphor.core.modeling.event import AttributeUpdated
from gaphor.core.modeling.properties import attribute
from gaphor.core.styling import CompiledStyleSheet, Style, StyleNode

SYSTEM_STYLE_SHEET = textwrap.dedent(
    """\
    * {
     background-color: transparent;
     color: black;
     font-family: sans;
     font-size: 14;
     line-width: 2;
     padding: 0;
    }

    *:drop {
     color: hsl(210, 100%, 25%);
     line-width: 3;
    }

    *:disabled {
     opacity: 0.5;
    }

    diagram {
     background-color: transparent;
    }

    dependency,
    interfacerealization {
        dash-style: 7 5;
    }

    dependency[on_folded_interface = true],
    interfacerealization[on_folded_interface = true] {
        dash-style: 0;
    }
    """
)

DEFAULT_STYLE_SHEET = textwrap.dedent(
    """\
    * {
     background-color: transparent;
     color: black;
     font-family: sans;
     font-size: 14;
    }

    diagram {
     /* line-style: sloppy 0.3; */
    }
    """
)


class StyleSheet(Element):
    _compiled_style_sheet: CompiledStyleSheet

    def __init__(self, id=None, model=None):
        super().__init__(id, model)

        self.compile_style_sheet()

    styleSheet: attribute[str] = attribute("styleSheet", str, DEFAULT_STYLE_SHEET)

    def compile_style_sheet(self) -> None:
        self._compiled_style_sheet = CompiledStyleSheet(
            SYSTEM_STYLE_SHEET, self.styleSheet
        )

    def match(self, node: StyleNode) -> Style:
        return self._compiled_style_sheet.match(node)

    def postload(self):
        super().postload()
        self.compile_style_sheet()

    def handle(self, event):
        # Ensure compiled style sheet is always up-to-date:
        if (
            isinstance(event, AttributeUpdated)
            and event.property is StyleSheet.styleSheet
        ):
            self.compile_style_sheet()

        super().handle(event)



from gaphor import UML
from gaphor.core import transactional
from gaphor.diagram.propertypages import (
    PropertyPageBase,
    PropertyPages,
    combo_box_text_auto_complete,
    new_resource_builder,
)

new_builder = new_resource_builder("gaphor.UML.deployments")


@PropertyPages.register(UML.Connector)
class InformationFlowPropertyPage(PropertyPageBase):
    """Information Flow on Connectors."""

    order = 30

    def __init__(self, subject: UML.Connector):
        super().__init__()
        self.subject = subject

    def construct(self):
        if not (
            self.subject
            and self.subject.informationFlow
            and type(self.subject.informationFlow[0]) is UML.InformationFlow
        ):
            return

        model = self.subject.model

        builder = new_builder(
            "information-flow-editor",
            signals={
                "information-flow-changed": (self._on_information_flow_changed,),
                "information-flow-combo-changed": (
                    self._on_information_flow_name_changed,
                ),
                "invert-direction-changed": (self._invert_direction_changed,),
            },
        )

        self.combo = combo = builder.get_object("information-flow-combo")
        use_flow = builder.get_object("use-information-flow")

        combo_box_text_auto_complete(
            combo, ((c.id, c.name) for c in model.select(UML.Class))
        )

        use_flow.set_active(
            self.subject.informationFlow
            and type(self.subject.informationFlow[0]) is UML.InformationFlow
        )
        self.combo.set_sensitive(use_flow.get_active())
        if self.subject.informationFlow and any(
            self.subject.informationFlow[:].conveyed
        ):
            combo.get_child().set_text(
                self.subject.informationFlow[0].conveyed[0].name or ""
            )

        return builder.get_object("information-flow-editor")

    @transactional
    def _on_information_flow_changed(self, switch, gparam):
        active = switch.get_active()
        subject = self.subject
        if active and not subject.informationFlow:
            iflow = subject.model.create(UML.InformationFlow)
            subject.informationFlow = iflow
            iflow.informationSource = subject.end[0].role
            iflow.informationTarget = subject.end[1].role
        elif not active and self.subject.informationFlow:
            self.subject.informationFlow[0].unlink()
        self.combo.set_sensitive(switch.get_active())
        self.combo.get_child().set_text("")

    @transactional
    def _on_information_flow_name_changed(self, combo):
        if self.subject.informationFlow:
            model = self.subject.model
            iflow = self.subject.informationFlow[0]
            if id := combo.get_active_id():
                iitem = model.lookup(id)
                assert isinstance(iitem, UML.Classifier)
                iflow.conveyed = iitem

    @transactional
    def _invert_direction_changed(self, button):
        if self.subject.informationFlow:
            iflow = self.subject.informationFlow[0]
            iflow.informationSource, iflow.informationTarget = (
                iflow.informationTarget,
                iflow.informationSource,
            )



"""The definition for the FTA section of the RAAML toolbox."""

from functools import partial

from gaphor.diagram.diagramtoolbox import ToolDef, ToolSection, new_item_factory
from gaphor.i18n import gettext
from gaphor.RAAML import diagramitems, raaml
from gaphor.UML import diagramitems as uml_items
from gaphor.UML.toolboxconfig import named_element_config, namespace_config

fta = ToolSection(
    gettext("Fault Tree Analysis"),
    (
        ToolDef(
            "dependency",
            gettext("Dependency"),
            "gaphor-dependency-symbolic",
            "<Shift>D",
            new_item_factory(uml_items.DependencyItem),
            handle_index=0,
        ),
        ToolDef(
            "and",
            gettext("AND Gate"),
            "gaphor-and-symbolic",
            "a",
            new_item_factory(
                diagramitems.ANDItem,
                raaml.AND,
                config_func=partial(namespace_config, name=gettext("AND")),
            ),
        ),
        ToolDef(
            "or",
            gettext("OR Gate"),
            "gaphor-or-symbolic",
            "o",
            new_item_factory(
                diagramitems.ORItem,
                raaml.OR,
                config_func=partial(namespace_config, name=gettext("OR")),
            ),
        ),
        ToolDef(
            "not",
            gettext("NOT Gate"),
            "gaphor-not-symbolic",
            "n",
            new_item_factory(
                diagramitems.NOTItem,
                raaml.NOT,
                config_func=partial(namespace_config, name=gettext("NOT")),
            ),
        ),
        ToolDef(
            "seq",
            gettext("Sequence Enforcing (SEQ) Gate"),
            "gaphor-seq-symbolic",
            "<Shift>S",
            new_item_factory(
                diagramitems.SEQItem,
                raaml.SEQ,
                config_func=partial(namespace_config, name=gettext("SEQ")),
            ),
        ),
        ToolDef(
            "xor",
            gettext("Exclusive OR Gate"),
            "gaphor-xor-symbolic",
            "x",
            new_item_factory(
                diagramitems.XORItem,
                raaml.XOR,
                config_func=partial(namespace_config, name=gettext("XOR")),
            ),
        ),
        ToolDef(
            "majority_vote",
            gettext("Majority Vote Gate"),
            "gaphor-majority_vote-symbolic",
            "m",
            new_item_factory(
                diagramitems.MajorityVoteItem,
                raaml.MAJORITY_VOTE,
                config_func=partial(
                    namespace_config, name=gettext("Majority Vote Gate")
                ),
            ),
        ),
        ToolDef(
            "inhibit",
            gettext("Inhibit Gate"),
            "gaphor-inhibit-symbolic",
            "i",
            new_item_factory(
                diagramitems.InhibitItem,
                raaml.INHIBIT,
                config_func=partial(namespace_config, name=gettext("Inhibit Gate")),
            ),
        ),
        ToolDef(
            "transfer-in",
            gettext("Transfer In"),
            "gaphor-transfer-in-symbolic",
            "t",
            new_item_factory(
                diagramitems.TransferInItem,
                raaml.TransferIn,
                config_func=partial(named_element_config, name=gettext("Transfer In")),
            ),
        ),
        ToolDef(
            "transfer-out",
            gettext("Transfer Out"),
            "gaphor-transfer-out-symbolic",
            "<Shift>T",
            new_item_factory(
                diagramitems.TransferOutItem,
                raaml.TransferOut,
                config_func=partial(namespace_config, name=gettext("Transfer Out")),
            ),
        ),
        ToolDef(
            "basic-event",
            gettext("Basic Event"),
            "gaphor-basic-event-symbolic",
            "<Shift>B",
            new_item_factory(
                diagramitems.BasicEventItem,
                raaml.BasicEvent,
                config_func=partial(namespace_config, name=gettext("Basic Event")),
            ),
        ),
        ToolDef(
            "conditional-event",
            gettext("Conditional Event"),
            "gaphor-conditional-event-symbolic",
            "c",
            new_item_factory(
                diagramitems.ConditionalEventItem,
                raaml.ConditionalEvent,
                config_func=partial(
                    namespace_config, name=gettext("Conditional Event")
                ),
            ),
        ),
        ToolDef(
            "undeveloped-event",
            gettext("Undeveloped Event"),
            "gaphor-undeveloped-event-symbolic",
            "<Shift>U",
            new_item_factory(
                diagramitems.UndevelopedEventItem,
                raaml.Undeveloped,
                config_func=partial(
                    named_element_config, name=gettext("Undeveloped Event")
                ),
            ),
        ),
        ToolDef(
            "dormant-event",
            gettext("Dormant Event"),
            "gaphor-dormant-event-symbolic",
            "d",
            new_item_factory(
                diagramitems.DormantEventItem,
                raaml.DormantEvent,
                config_func=partial(namespace_config, name=gettext("Dormant Event")),
            ),
        ),
        ToolDef(
            "house-event",
            gettext("House Event"),
            "gaphor-house-event-symbolic",
            "h",
            new_item_factory(
                diagramitems.HouseEventItem,
                raaml.HouseEvent,
                config_func=partial(namespace_config, name=gettext("House Event")),
            ),
        ),
        ToolDef(
            "zero-event",
            gettext("Zero Event"),
            "gaphor-zero-event-symbolic",
            "z",
            new_item_factory(
                diagramitems.ZeroEventItem,
                raaml.ZeroEvent,
                config_func=partial(namespace_config, name=gettext("Zero Event")),
            ),
        ),
        ToolDef(
            "top-event",
            gettext("Top Event"),
            "gaphor-top-event-symbolic",
            "p",
            new_item_factory(
                diagramitems.TopEventItem,
                raaml.TopEvent,
                config_func=partial(namespace_config, name=gettext("Top Event")),
            ),
        ),
        ToolDef(
            "intermediate-event",
            gettext("Intermediate Event"),
            "gaphor-intermediate-event-symbolic",
            "<Shift>I",
            new_item_factory(
                diagramitems.IntermediateEventItem,
                raaml.IntermediateEvent,
                config_func=partial(
                    namespace_config, name=gettext("Intermediate Event")
                ),
            ),
        ),
    ),
)



"""UI related events."""


class ElementOpened:
    def __init__(self, element):
        self.element = element


class DiagramOpened:
    def __init__(self, diagram):
        self.diagram = diagram


class DiagramClosed:
    def __init__(self, diagram):
        self.diagram = diagram


class CurrentDiagramChanged:
    def __init__(self, diagram):
        self.diagram = diagram


class DiagramSelectionChanged:
    def __init__(self, diagram_view, focused_item, selected_items):
        self.diagram_view = diagram_view
        self.focused_item = focused_item
        self.selected_items = selected_items


class ToolSelected:
    def __init__(self, tool_name):
        self.tool_name = tool_name



"""The UML Modeling Language module is the entrypoint for UML related
assets."""

from typing import Iterable

from gaphor.abc import ModelingLanguage
from gaphor.core import gettext
from gaphor.diagram.diagramtoolbox import DiagramType, ToolboxDefinition
from gaphor.UML import diagramitems, uml
from gaphor.UML.toolbox import uml_diagram_types, uml_toolbox_actions


class UMLModelingLanguage(ModelingLanguage):
    @property
    def name(self) -> str:
        return gettext("UML")

    @property
    def toolbox_definition(self) -> ToolboxDefinition:
        return uml_toolbox_actions

    @property
    def diagram_types(self) -> Iterable[DiagramType]:
        """Return an iterator (id, name) for each diagram type."""
        yield from uml_diagram_types

    def lookup_element(self, name):
        return getattr(uml, name, None) or getattr(diagramitems, name, None)



import logging

from gaphor.storage.xmlwriter import XMLWriter

logger = logging.getLogger(__name__)


class XMIExport:

    XMI_VERSION = "2.1"
    XMI_NAMESPACE = "http://schema.omg.org/spec/XMI/2.1"
    UML_NAMESPACE = "http://schema.omg.org/spec/UML/2.1"
    XMI_PREFIX = "XMI"
    UML_PREFIX = "UML"

    def __init__(self, element_factory):
        self.element_factory = element_factory
        self.handled_ids = []

    def handle(self, xmi, element):
        logger.debug(f"Handling {element.__class__.__name__}")
        try:
            handler_name = f"handle{element.__class__.__name__}"
            handler = getattr(self, handler_name)
            idref = element.id in self.handled_ids
            handler(xmi, element, idref=idref)
            if not idref:
                self.handled_ids.append(element.id)
        except AttributeError as e:
            logger.warning(f"Missing handler for {element.__class__.__name__}:{e}")
        except Exception as e:
            logger.error(f"Failed to handle {element.__class__.__name__}:{e}")

    def handlePackage(self, xmi, element, idref=False):

        attributes = {
            f"{self.XMI_PREFIX}:id": element.id,
            "name": element.name,
            "visibility": element.visibility,
        }

        xmi.startElement(f"{self.UML_PREFIX}:Package", attrs=attributes)

        for ownedMember in element.ownedMember:
            xmi.startElement("ownedMember", attrs={})
            self.handle(xmi, ownedMember)
            xmi.endElement("ownedMember")

        xmi.endElement(f"{self.UML_PREFIX}:Package")

    def handleClass(self, xmi, element, idref=False):

        attributes = {}

        if idref:
            attributes[f"{self.XMI_PREFIX}:idref"] = element.id
        else:
            attributes[f"{self.XMI_PREFIX}:id"] = element.id
            attributes["name"] = element.name
            attributes["isAbstract"] = str(element.isAbstract)

        xmi.startElement(f"{self.UML_PREFIX}:Class", attrs=attributes)

        if not idref:

            for ownedAttribute in element.ownedAttribute:
                xmi.startElement("ownedAttribute", attrs={})
                self.handle(xmi, ownedAttribute)
                xmi.endElement("ownedAttribute")

            for ownedOperation in element.ownedOperation:
                xmi.startElement("ownedOperation", attrs={})
                self.handle(xmi, ownedOperation)
                xmi.endElement("ownedOperation")

        xmi.endElement(f"{self.UML_PREFIX}:Class")

    def handleProperty(self, xmi, element, idref=False):

        attributes = {
            f"{self.XMI_PREFIX}:id": element.id,
            "isStatic": str(element.isStatic),
            "isOrdered": str(element.isOrdered),
            "isUnique": str(element.isUnique),
            "isDerived": str(element.isDerived),
            "isDerivedUnion": str(element.isDerivedUnion),
            "isReadOnly": str(element.isReadOnly),
        }

        if element.name is not None:
            attributes["name"] = element.name

        xmi.startElement(f"{self.UML_PREFIX}:Property", attrs=attributes)

        # TODO: This should be type, not typeValue.
        if element.typeValue is not None:
            xmi.startElement("type", attrs={})
            self.handle(xmi, element.typeValue)
            xmi.endElement("type")

        xmi.endElement(f"{self.UML_PREFIX}:Property")

    def handleOperation(self, xmi, element, idref=False):

        attributes = {
            f"{self.XMI_PREFIX}:id": element.id,
            "isStatic": str(element.isStatic),
            "isQuery": str(element.isQuery),
            "name": element.name,
        }

        xmi.startElement(f"{self.XMI_PREFIX}:Operation", attrs=attributes)

        for ownedParameter in element.ownedParameter:
            xmi.startElement("ownedElement", attrs={})
            self.handle(xmi, ownedParameter)
            xmi.endElement("ownedElement")

        xmi.endElement(f"{self.XMI_PREFIX}:Operation")

    def handleParameter(self, xmi, element, idref=False):

        attributes = {
            f"{self.XMI_PREFIX}:id": element.id,
            "isOrdered": str(element.isOrdered),
            "isUnique": str(element.isUnique),
            "direction": element.direction,
            "name": element.name,
        }

        xmi.startElement(f"{self.XMI_PREFIX}:Parameter", attrs=attributes)

        xmi.endElement(f"{self.XMI_PREFIX}:Parameter")

    def handleLiteralSpecification(self, xmi, element, idref=False):

        attributes = {f"{self.XMI_PREFIX}:id": element.id, "value": element.value}
        xmi.startElement(f"{self.UML_PREFIX}:LiteralSpecification", attrs=attributes)

        xmi.endElement(f"{self.UML_PREFIX}:LiteralSpecification")

    def handleAssociation(self, xmi, element, idref=False):

        attributes = {
            f"{self.XMI_PREFIX}:id": element.id,
            "isDerived": str(element.isDerived),
        }

        xmi.startElement(f"{self.UML_PREFIX}:Association", attrs=attributes)

        for memberEnd in element.memberEnd:
            xmi.startElement("memberEnd", attrs={})
            self.handle(xmi, memberEnd)
            xmi.endElement("memberEnd")

        for ownedEnd in element.ownedEnd:
            xmi.startElement("ownedEnd", attrs={})
            self.handle(xmi, ownedEnd)
            xmi.endElement("ownedEnd")

        xmi.endElement(f"{self.UML_PREFIX}:Association")

    def handleDependency(self, xmi, element, idref=False, name="Dependency"):

        attributes = {f"{self.XMI_PREFIX}:id": element.id}
        xmi.startElement(f"{self.UML_PREFIX}:{name}", attrs=attributes)

        for client in element.client:
            xmi.startElement("client", attrs={})
            self.handle(xmi, client)
            xmi.endElement("client")

        for supplier in element.supplier:
            xmi.startElement("supplier", attrs={})
            self.handle(xmi, supplier)
            xmi.endElement("supplier")

        xmi.endElement(f"{self.UML_PREFIX}:{name}")

    def handleGeneralization(self, xmi, element, idref=False):

        attributes = {
            f"{self.XMI_PREFIX}:id": element.id,
            "isSubstitutable": str(element.isSubstitutable),
        }

        xmi.startElement(f"{self.UML_PREFIX}:Generalization", attrs=attributes)

        if element.general:
            xmi.startElement("general", attrs={})
            self.handle(xmi, element.general)
            xmi.endElement("general")

        if element.specific:
            xmi.startElement("specific", attrs={})
            self.handle(xmi, element.specific)
            xmi.endElement("specific")

        xmi.endElement(f"{self.UML_PREFIX}:Generalization")

    def handleRealization(self, xmi, element, idref=False):
        self.handleDependency(xmi, element, idref, name="Realization")

    def handleInterface(self, xmi, element, idref=False):

        attributes = {f"{self.XMI_PREFIX}:id": element.id}
        xmi.startElement(f"{self.UML_PREFIX}:Interface", attrs=attributes)

        for ownedAttribute in element.ownedAttribute:
            xmi.startElement("ownedAttribute", attrs={})
            self.handle(xmi, ownedAttribute)
            xmi.endElement("ownedAttribute")

        for ownedOperation in element.ownedOperation:
            xmi.startElement("ownedOperation", attrs={})
            self.handle(xmi, ownedOperation)
            xmi.endElement("ownedOperation")

        xmi.endElement(f"{self.UML_PREFIX}:Interface")

    def handleDiagram(self, xmi, element, idref=False):
        pass

    def export(self, filename):
        with open(filename, "w", encoding="utf-8") as out:
            xmi = XMLWriter(out)

            attributes = {
                "xmi.version": self.XMI_VERSION,
                "xmlns:xmi": self.XMI_NAMESPACE,
                "xmlns:UML": self.UML_NAMESPACE,
            }

            xmi.startElement("XMI", attrs=attributes)

            for package in self.element_factory.select(self.select_package):
                self.handle(xmi, package)

            for generalization in self.element_factory.select(
                self.select_generalization
            ):
                self.handle(xmi, generalization)

            for realization in self.element_factory.select(self.select_realization):
                self.handle(xmi, realization)

            xmi.endElement("XMI")

            logger.debug(self.handled_ids)

    def select_package(self, element):
        return element.__class__.__name__ == "Package"

    def select_generalization(self, element):
        return element.__class__.__name__ == "Generalization"

    def select_realization(self, element):
        return element.__class__.__name__ == "InterfaceRealization"



from gaphas.handlemove import HandleMove, ItemHandleMove
from gaphas.tool import item_tool as _item_tool
from gaphas.types import Pos
from gaphas.view import GtkView
from gi.repository import Gtk

from gaphor.core.modeling import Presentation
from gaphor.diagram.connectors import ItemTemporaryDisconnected
from gaphor.diagram.presentation import LinePresentation


def item_tool(view: GtkView) -> Gtk.GestureDrag:
    gesture = _item_tool(view)
    gesture.connect_after("drag-end", on_drag_end)
    return gesture


def on_drag_end(gesture, _offset_x, _offset_y):
    view = gesture.get_widget()
    view.selection.dropzone_item = None


@HandleMove.register(Presentation)
@HandleMove.register(LinePresentation)
class PresentationHandleMove(ItemHandleMove):
    def start_move(self, pos: Pos) -> None:
        super().start_move(pos)
        model = self.view.model
        assert model
        if cinfo := model.connections.get_connection(self.handle):
            self.item.handle(
                ItemTemporaryDisconnected(
                    cinfo.item, cinfo.handle, cinfo.connected, cinfo.port
                )
            )



from gaphor.C4Model import c4model
from gaphor.core import gettext
from gaphor.core.styling import FontWeight, JustifyContent, TextAlign
from gaphor.diagram.presentation import ElementPresentation, Named
from gaphor.diagram.shapes import Box, Text, ellipse, stroke
from gaphor.diagram.support import represents


@represents(c4model.C4Database)
class C4DatabaseItem(Named, ElementPresentation):
    def __init__(self, diagram, id=None):
        super().__init__(diagram, id)

        self.watch("subject[NamedElement].name")
        self.watch("subject[C4Container].technology")
        self.watch("subject[C4Container].description")
        self.watch("subject[C4Container].type")

    def update_shapes(self, event=None):
        self.shape = Box(
            Box(
                Text(
                    text=lambda: self.subject.name or "",
                    style={"font-weight": FontWeight.BOLD},
                ),
                Text(
                    text=lambda: self.subject.technology
                    and f"[{gettext(self.subject.type)}: {self.subject.technology}]"
                    or f"[{gettext(self.subject.type)}]",
                    style={"font-size": "x-small"},
                ),
                Text(
                    text=lambda: self.subject.description or "",
                    width=lambda: self.width - 8,
                    style={"padding": (4, 0, 0, 0)},
                ),
                style={"padding": (4, 4, 4, 4)},
            ),
            style={
                "text-align": TextAlign.LEFT
                if self.diagram and self.children
                else TextAlign.CENTER,
                "justify-content": JustifyContent.END
                if self.diagram and self.children
                else JustifyContent.CENTER,
            },
            draw=draw_database,
        )


def draw_database(box, context, bounding_box):
    cr = context.cairo
    d = 0.38
    x, y, width, height = bounding_box

    x1 = width + x
    y1 = height + y
    ellipse(cr, x, y - height * d / 2, width, height * d)
    cr.move_to(x1, y)
    cr.line_to(x1, y1)
    cr.curve_to(x1, y1, width / 2, y1 + height * d, x, y1)
    cr.line_to(x, y)

    stroke(context)



import abc
from typing import Tuple

from gaphor.abc import Service


class UIComponent(Service):
    """A user interface component."""

    ui_name: str  # "The UIComponent name, provided by the loader"

    size: Tuple[int, int]  # "Size used for floating the component"

    @abc.abstractmethod
    def open(self):
        """Create and display the UI components (windows)."""

    @abc.abstractmethod
    def close(self):
        """Close the UI component.

        The component can decide to hide or destroy the UI components.
        """

    def shutdown(self):
        """Shut down this component.

        It's not supposed to be opened again.
        """
        self.close()



from gaphor import UML
from gaphor.core import gettext
from gaphor.diagram.presentation import AttachedPresentation, Named
from gaphor.diagram.shapes import (
    Box,
    IconBox,
    Text,
    TextAlign,
    VerticalAlign,
    draw_border,
)
from gaphor.diagram.support import represents
from gaphor.UML.recipes import stereotypes_str


def text_position(position):
    return {
        "text-align": TextAlign.LEFT if position == "left" else TextAlign.RIGHT,
        "vertical-align": VerticalAlign.BOTTOM
        if position == "bottom"
        else VerticalAlign.TOP,
    }


class PinItem(Named, AttachedPresentation[UML.Pin]):
    def __init__(self, diagram, id=None):
        super().__init__(diagram, id, width=16, height=16)
        self.watch("subject[NamedElement].name")

    def pin_type(self):
        return ""

    def update_shapes(self):
        position = self.connected_side()
        self.update_width(
            self.width,
            factor=0 if position == "left" else 1 if position == "right" else 0.5,
        )
        self.update_height(
            self.width,
            factor=0 if position == "top" else 1 if position == "bottom" else 0.5,
        )

        self.shape = IconBox(
            Box(style={"background-color": (1, 1, 1, 1)}, draw=draw_border),
            Text(
                text=lambda: stereotypes_str(
                    self.subject, [] if self.subject else [self.pin_type()]
                )
            ),
            Text(text=lambda: self.subject and self.subject.name or ""),
            style=text_position(self.connected_side()),
        )


@represents(UML.InputPin)
class InputPinItem(PinItem):
    def pin_type(self):
        return gettext("input pin")


@represents(UML.OutputPin)
class OutputPinItem(PinItem):
    def pin_type(self):
        return gettext("output pin")



"""Use case diagram item."""

from gaphor import UML
from gaphor.diagram.presentation import Classified, ElementPresentation
from gaphor.diagram.shapes import Box, Text, draw_ellipse
from gaphor.diagram.support import represents
from gaphor.diagram.text import FontStyle, FontWeight
from gaphor.UML.recipes import stereotypes_str


@represents(UML.UseCase)
class UseCaseItem(Classified, ElementPresentation):
    """Presentation of gaphor.UML.UseCase."""

    def __init__(self, diagram, id=None):
        super().__init__(diagram, id, height=30)

        self.watch("subject[NamedElement].name")
        self.watch("subject.appliedStereotype.classifier.name")
        self.watch("subject[Classifier].isAbstract", self.update_shapes)

    def update_shapes(self, event=None):
        self.shape = Box(
            Text(
                text=lambda: stereotypes_str(self.subject),
            ),
            Text(
                text=lambda: self.subject.name or "",
                width=lambda: self.width,
                style={
                    "font-weight": FontWeight.BOLD,
                    "font-style": FontStyle.ITALIC
                    if self.subject and self.subject.isAbstract
                    else FontStyle.NORMAL,
                },
            ),
            draw=draw_ellipse,
        )



"""The main application window."""

from __future__ import annotations

import importlib.resources
import logging
from pathlib import Path

from gi.repository import Gio, GLib, Gtk

from gaphor.abc import ActionProvider, Service
from gaphor.core import event_handler, gettext
from gaphor.event import (
    ActionEnabled,
    ActiveSessionChanged,
    ModelLoaded,
    ModelSaved,
    SessionCreated,
    SessionShutdownRequested,
)
from gaphor.i18n import translated_ui_string
from gaphor.services.modelinglanguage import ModelingLanguageChanged
from gaphor.services.undomanager import UndoManagerStateChanged
from gaphor.ui.abc import UIComponent
from gaphor.ui.actiongroup import window_action_group
from gaphor.ui.layout import deserialize, is_maximized
from gaphor.ui.notification import InAppNotifier

log = logging.getLogger(__name__)


def new_builder():
    builder = Gtk.Builder()
    ui_file = "mainwindow.glade" if Gtk.get_major_version() == 3 else "mainwindow.ui"
    builder.add_from_string(translated_ui_string("gaphor.ui", ui_file))
    return builder


def create_hamburger_model(export_menu, tools_menu):
    model = Gio.Menu.new()

    part = Gio.Menu.new()
    part.append(gettext("Open a Model…"), "app.file-open")
    part.append(gettext("Recent Models…"), "app.recent-files")
    part.append(gettext("New Model…"), "app.new-model")
    model.append_section(None, part)

    part = Gio.Menu.new()
    part.append(gettext("Save"), "win.file-save")
    part.append(gettext("Save As…"), "win.file-save-as")
    part.append_submenu(gettext("Export"), export_menu)
    model.append_section(None, part)

    part = Gio.Menu.new()
    part.append_submenu(gettext("Tools"), tools_menu)
    model.append_section(None, part)

    part = Gio.Menu.new()
    part.append(gettext("Keyboard Shortcuts"), "win.shortcuts")
    part.append(gettext("About Gaphor"), "win.about")
    model.append_section(None, part)

    return model


def create_modeling_language_model(modeling_language):
    model = Gio.Menu.new()
    for id, name in modeling_language.modeling_languages:
        menu_item = Gio.MenuItem.new(name, "win.select-modeling-language")
        menu_item.set_attribute_value("target", GLib.Variant.new_string(id))
        model.append_item(menu_item)
    return model


def create_diagram_types_model(modeling_language):
    model = Gio.Menu.new()

    part = Gio.Menu.new()
    for id, name, _ in modeling_language.diagram_types:
        menu_item = Gio.MenuItem.new(name, "win.create-diagram")
        menu_item.set_attribute_value("target", GLib.Variant.new_string(id))
        part.append_item(menu_item)
    model.append_section(None, part)

    part = Gio.Menu.new()
    menu_item = Gio.MenuItem.new(gettext("New Generic Diagram"), "win.create-diagram")
    menu_item.set_attribute_value("target", GLib.Variant.new_string(""))
    part.append_item(menu_item)
    model.append_section(None, part)

    return model


def popup_set_model(popup, model):
    if Gtk.get_major_version() == 3:
        popup.bind_model(model, None)
    else:
        popup.set_menu_model(model)


class MainWindow(Service, ActionProvider):
    """The main window for the application.

    It contains a Namespace-based tree view and a menu and a statusbar.
    """

    size = property(lambda s: s.properties.get("ui.window-size", (860, 580)))

    def __init__(
        self,
        event_manager,
        component_registry,
        properties,
        modeling_language,
        export_menu,
        tools_menu,
    ):
        self.event_manager = event_manager
        self.component_registry = component_registry
        self.properties = properties
        self.modeling_language = modeling_language
        self.export_menu = export_menu
        self.tools_menu = tools_menu

        self.window: Gtk.Window = None
        self.action_group: Gio.ActionGroup = None
        self.title: Gtk.Label = None
        self.subtitle: Gtk.Label = None
        self.filename: Path | None = None
        self.model_changed = False
        self.layout = None
        self.modeling_language_name = None
        self.diagram_types = None
        self.in_app_notifier = None

        event_manager.subscribe(self._on_file_manager_state_changed)

    def shutdown(self):
        if self.window:
            self.window.destroy()
            self.window = None

        em = self.event_manager
        em.unsubscribe(self._on_file_manager_state_changed)
        em.unsubscribe(self._on_undo_manager_state_changed)
        em.unsubscribe(self._on_action_enabled)
        em.unsubscribe(self._on_modeling_language_selection_changed)
        if self.in_app_notifier:
            em.unsubscribe(self.in_app_notifier.handle)
            self.in_app_notifier = None

    def get_ui_component(self, name):
        return self.component_registry.get(UIComponent, name)

    def open(self, gtk_app=None):
        """Open the main window."""

        builder = new_builder()
        self.window = builder.get_object("main-window")
        self.window.set_application(gtk_app)

        select_modeling_language = builder.get_object("select-modeling-language")
        popup_set_model(
            select_modeling_language,
            create_modeling_language_model(self.modeling_language),
        )
        self.modeling_language_name = builder.get_object("modeling-language-name")

        self.diagram_types = builder.get_object("diagram-types")
        popup_set_model(
            self.diagram_types, create_diagram_types_model(self.modeling_language)
        )

        hamburger = builder.get_object("hamburger")
        popup_set_model(
            hamburger,
            create_hamburger_model(self.export_menu.menu, self.tools_menu.menu),
        )

        self.title = builder.get_object("title")
        self.subtitle = builder.get_object("subtitle")
        self.set_title()

        self.window.set_default_size(*self.size)

        def _factory(name):
            comp = self.get_ui_component(name)
            return comp.open()

        main_content = builder.get_object("main-content")
        self.layout = deserialize(
            main_content,
            (importlib.resources.files("gaphor.ui") / "layout.xml").read_text(),
            _factory,
            self.properties,
        )

        self.action_group, shortcuts = window_action_group(self.component_registry)
        self.window.insert_action_group("win", self.action_group)

        if Gtk.get_major_version() == 3:
            self.window.add_accel_group(shortcuts)
        else:
            self.window.add_controller(Gtk.ShortcutController.new_for_model(shortcuts))

        self._on_modeling_language_selection_changed()

        self.window.set_resizable(True)
        if Gtk.get_major_version() == 3:
            self.window.show_all()
            self.window.connect("delete-event", self._on_window_close_request)
            self.window.connect("size-allocate", self._on_window_size_allocate)
        else:
            self.window.connect("close-request", self._on_window_close_request)
            self.window.connect("notify::default-height", self._on_window_size_changed)
            self.window.connect("notify::default-width", self._on_window_size_changed)
            self.window.show()

        self.window.connect("notify::is-active", self._on_window_active)

        self.in_app_notifier = InAppNotifier(builder)
        em = self.event_manager
        em.subscribe(self._on_undo_manager_state_changed)
        em.subscribe(self._on_action_enabled)
        em.subscribe(self._on_modeling_language_selection_changed)
        em.subscribe(self.in_app_notifier.handle)

    def set_title(self):
        """Sets the window title."""
        if not self.window:
            return

        if self.filename:
            p = self.filename
            title = p.stem
            subtitle = str(p).replace(str(Path.home()), "~")
        else:
            title = "Gaphor"
            subtitle = gettext("New model")
        if self.model_changed:
            title += " [" + gettext("edited") + "]"
        self.title.set_text(title)
        self.subtitle.set_text(subtitle)
        self.window.set_title(title)

    # Signal callbacks:

    @event_handler(SessionCreated, ModelLoaded, ModelSaved)
    def _on_file_manager_state_changed(
        self, event: SessionCreated | ModelLoaded | ModelSaved
    ):
        self.model_changed = False
        self.filename = Path(event.filename) if event.filename else None
        self.set_title()
        if self.window:
            self.window.present()

    @event_handler(UndoManagerStateChanged)
    def _on_undo_manager_state_changed(self, event):
        undo_manager = event.service
        if self.model_changed != undo_manager.can_undo():
            self.model_changed = undo_manager.can_undo()
            self.set_title()

    @event_handler(ActionEnabled)
    def _on_action_enabled(self, event):
        if self.action_group and event.scope == "win":
            a = self.action_group.lookup_action(event.name)
            a.set_enabled(event.enabled)

    @event_handler(ModelingLanguageChanged)
    def _on_modeling_language_selection_changed(self, event=None):
        if self.modeling_language_name:
            self.modeling_language_name.set_label(
                gettext("Profile: {name}").format(name=self.modeling_language.name)
            )
        if self.diagram_types:
            popup_set_model(
                self.diagram_types, create_diagram_types_model(self.modeling_language)
            )

    def _on_window_active(self, window, prop):
        self.event_manager.handle(ActiveSessionChanged(self))

    def _on_window_close_request(self, window, event=None):
        self.event_manager.handle(SessionShutdownRequested(self))
        return True

    if Gtk.get_major_version() == 3:

        def _on_window_size_allocate(self, window, allocation):
            if not is_maximized(window):
                width, height = window.get_size()
                self.properties.set("ui.window-size", (width, height))

    else:

        def _on_window_size_changed(self, window, gspec):
            if not is_maximized(window):
                width, height = window.get_default_size()
                self.properties.set("ui.window-size", (width, height))



# flake8: noqa F401
from gi.repository import Gtk

if Gtk.get_major_version() == 3:
    from gaphor.ui.namespace import Namespace as TreeComponent
else:
    from gaphor.ui.treecomponent import TreeComponent  # type: ignore[assignment]



from gi.repository import Gdk, Gtk


def widget_add_hover_support(widget):
    """Add hover support to a widget.

    Gtk3 only.
    """
    widget.add_events(
        Gdk.EventMask.ENTER_NOTIFY_MASK
        | Gdk.EventMask.LEAVE_NOTIFY_MASK
        | Gdk.EventMask.POINTER_MOTION_MASK
    )

    def hover(widget, event):
        widget.set_state_flags(Gtk.StateFlags.PRELIGHT, False)

    def unhover(widget, event):
        widget.unset_state_flags(Gtk.StateFlags.PRELIGHT)

    widget.connect("motion-notify-event", hover)
    widget.connect("leave-notify-event", unhover)


def flowbox_add_hover_support(flowbox):
    """Add hover support to a flowbox.

    Gtk3 only.
    """
    flowbox.add_events(
        Gdk.EventMask.ENTER_NOTIFY_MASK
        | Gdk.EventMask.LEAVE_NOTIFY_MASK
        | Gdk.EventMask.POINTER_MOTION_MASK
    )

    hover_child: Gtk.Widget = None

    def hover(widget, event):
        nonlocal hover_child
        child = widget.get_child_at_pos(event.x, event.y)
        if hover_child and child is not hover_child:
            hover_child.unset_state_flags(Gtk.StateFlags.PRELIGHT)
        if child:
            child.set_state_flags(Gtk.StateFlags.PRELIGHT, False)
        hover_child = child

    def unhover(widget, event):
        if hover_child:
            hover_child.unset_state_flags(Gtk.StateFlags.PRELIGHT)

    flowbox.connect("motion-notify-event", hover)
    flowbox.connect("leave-notify-event", unhover)



import os

import gi

if os.getenv("GAPHOR_USE_GTK") != "NONE":
    # Allow to explicitly *not* initialize GTK (for docs, mainly)
    gtk_version = "3.0" if os.getenv("GAPHOR_USE_GTK") == "3" else "4.0"

    gi.require_version("Gtk", gtk_version)
    gi.require_version("Gdk", gtk_version)



"""Stereotype property page."""
from gi.repository import Gtk

from gaphor import UML
from gaphor.core import transactional
from gaphor.core.modeling.element import Element
from gaphor.diagram.propertypages import PropertyPageBase, PropertyPages
from gaphor.UML.profiles.metaclasspropertypage import new_builder


@PropertyPages.register(UML.Element)
class StereotypePage(PropertyPageBase):

    order = 40

    def __init__(self, item):
        self.item = item

    def construct(self):

        subject = self.item.subject
        if not subject:
            return None

        stereotypes = UML.recipes.get_stereotypes(subject)
        if not stereotypes:
            return None

        model, toggle_stereotype_handler, set_slot_value_handler = stereotype_model(
            subject
        )

        builder = new_builder(
            "stereotypes-editor",
            signals={
                "show-stereotypes-changed": (self._on_show_stereotypes_change,),
                "toggle-stereotype": toggle_stereotype_handler,
                "set-slot-value": set_slot_value_handler,
            },
        )

        show_stereotypes = builder.get_object("show-stereotypes")

        if hasattr(self.item, "show_stereotypes"):
            show_stereotypes.set_active(self.item.show_stereotypes)
        elif Gtk.get_major_version() == 3:
            show_stereotypes.destroy()
        else:
            show_stereotypes.unparent()

        stereotype_list = builder.get_object("stereotype-list")
        stereotype_list.set_model(model)

        return builder.get_object("stereotypes-editor")

    @transactional
    def _on_show_stereotypes_change(self, button, gparam):
        self.item.show_stereotypes = button.get_active()


def stereotype_model(subject):
    model = Gtk.TreeStore.new(
        [
            str,  # stereotype/attribute
            str,  # value
            bool,  # active / is applied stereotype
            bool,  # visible  checkbox (is stereotype)
            bool,  # value editable
            object,  # stereotype / attribute
            object,  # instance specification
            object,  # slot
        ]
    )
    refresh(subject, model)

    return model, (toggle_stereotype, subject, model), (set_value, model)


def refresh(subject: Element, model: Gtk.TreeStore):
    stereotypes = UML.recipes.get_stereotypes(subject)
    instances = subject.appliedStereotype

    def upsert(path, parent, row_data):
        try:
            new_row = model.get_iter(path)
        except ValueError:
            new_row = model.append(parent, row_data)
        else:
            row = model[path]
            row[:] = row_data
        return new_row

    for st_index, st in enumerate(stereotypes):
        applied = next(
            (applied for applied in instances if st in applied.classifier), None
        )

        parent = upsert(
            f"{st_index}",
            None,
            (st.name, "", bool(applied), True, False, st, None, None),
        )
        for attr_index, attr in enumerate(all_attributes(st)):
            slot = (
                next(
                    (slot for slot in applied.slot if slot.definingFeature is attr),
                    None,
                )
                if applied
                else None
            )
            value = slot.value if slot else ""
            upsert(
                f"{st_index}:{attr_index}",
                parent,
                (
                    attr.name,
                    value,
                    bool(applied),
                    False,
                    bool(applied),
                    attr,
                    applied,
                    slot,
                ),
            )


def all_attributes(stereotype, seen=None):
    if seen is None:
        seen = {stereotype}

    for super_type in stereotype.generalization[:].general[:]:
        if super_type not in seen:
            seen.add(super_type)
            yield from all_attributes(super_type)
    yield from (attr for attr in stereotype.ownedAttribute if not attr.association)


@transactional
def toggle_stereotype(renderer, path, subject, model):
    row = model[path]
    _, _, is_applied, _, _, stereotype, _, _ = row
    value = not is_applied

    if value:
        UML.recipes.apply_stereotype(subject, stereotype)
    else:
        UML.recipes.remove_stereotype(subject, stereotype)

    row[2] = value

    refresh(subject, model)


@transactional
def set_value(renderer, path, value, model):
    """Set value of stereotype property applied to a UML element."""
    row = model[path]
    _, _, _, _, _, attr, applied, slot = row
    if isinstance(attr, UML.Stereotype):
        return  # don't edit stereotype rows

    if slot is None:
        if value:
            slot = UML.recipes.add_slot(applied, attr)

        else:
            return  # nothing to do and don't create slot without value

    assert slot

    if value:
        slot.value = value
    else:
        # no value, then remove slot
        del applied.slot[slot]
        slot = None
        value = ""

    row[1] = value
    row[7] = slot



"""The definition for the requirements section of the toolbox."""

from gaphas.item import SE

from gaphor.core import gettext
from gaphor.diagram.diagramtoolbox import ToolDef, ToolSection, new_item_factory
from gaphor.SysML import diagramitems as sysml_items
from gaphor.SysML import sysml
from gaphor.UML.classes import ContainmentItem
from gaphor.UML.toolboxconfig import namespace_config

requirements = ToolSection(
    gettext("Requirements"),
    (
        ToolDef(
            "toolbox-requirement",
            gettext("Requirement"),
            "gaphor-requirement-symbolic",
            "r",
            new_item_factory(
                sysml_items.RequirementItem,
                sysml.Requirement,
                config_func=namespace_config,
            ),
            handle_index=SE,
        ),
        ToolDef(
            "toolbox-satisfy-dependency",
            gettext("Satisfy"),
            "gaphor-satisfy-symbolic",
            "<Shift>I",
            new_item_factory(sysml_items.SatisfyItem),
        ),
        ToolDef(
            "toolbox-derive-reqt-dependency",
            gettext("Derive Requirement"),
            "gaphor-derive-symbolic",
            "<Shift>D",
            new_item_factory(sysml_items.DeriveReqtItem),
        ),
        ToolDef(
            "toolbox-trace-dependency",
            gettext("Trace"),
            "gaphor-trace-symbolic",
            "y",
            new_item_factory(sysml_items.TraceItem),
        ),
        ToolDef(
            "toolbox-refine-dependency",
            gettext("Refine"),
            "gaphor-refine-symbolic",
            None,
            new_item_factory(sysml_items.RefineItem),
        ),
        ToolDef(
            "toolbox-verify-dependency",
            gettext("Verify"),
            "gaphor-verify-symbolic",
            "<Shift>V",
            new_item_factory(sysml_items.VerifyItem),
        ),
        ToolDef(
            "toolbox-containment",
            gettext("Containment"),
            "gaphor-containment-symbolic",
            "<Shift>M",
            new_item_factory(ContainmentItem),
        ),
    ),
)



"""The action definition for the RAAML toolbox."""
from gaphor.diagram.diagramtoolbox import (
    DiagramType,
    DiagramTypes,
    ToolboxDefinition,
    general_tools,
)
from gaphor.i18n import gettext
from gaphor.RAAML.fta.ftatoolbox import fta
from gaphor.RAAML.stpa.stpatoolbox import stpa

raaml_toolbox_actions: ToolboxDefinition = (
    general_tools,
    fta,
    stpa,
)

raaml_diagram_types: DiagramTypes = (
    DiagramType("fta", gettext("New FTA Diagram"), (fta,)),
    DiagramType("stpa", gettext("New STPA Diagram"), (stpa,)),
)



"""The definition for the STPA section of the RAAML toolbox."""

from gaphor.diagram.diagramtoolbox import ToolDef, ToolSection, new_item_factory
from gaphor.i18n import gettext
from gaphor.RAAML import diagramitems, raaml
from gaphor.SysML import diagramitems as sysml_items
from gaphor.UML import diagramitems as uml_items
from gaphor.UML.toolboxconfig import (
    default_namespace,
    named_element_config,
    namespace_config,
)


def loss_config(new_item):
    default_namespace(new_item)
    new_item.subject.name = gettext("Loss")


def hazard_config(new_item):
    default_namespace(new_item)
    new_item.subject.name = gettext("Hazard")


def abstract_operational_situation_config(new_item):
    default_namespace(new_item)
    new_item.subject.name = gettext("Abstract Operational Situation")


stpa = ToolSection(
    "STPA",
    (
        ToolDef(
            "toolbox-generalization",
            gettext("Generalization"),
            "gaphor-generalization-symbolic",
            "<Shift>G",
            new_item_factory(uml_items.GeneralizationItem),
            handle_index=0,
        ),
        ToolDef(
            "loss",
            gettext("Loss"),
            "gaphor-loss-symbolic",
            "<Shift>L",
            new_item_factory(
                sysml_items.BlockItem, raaml.Loss, config_func=loss_config
            ),
        ),
        ToolDef(
            "hazard",
            gettext("Hazard"),
            "gaphor-hazard-symbolic",
            "<Shift>H",
            new_item_factory(
                sysml_items.BlockItem, raaml.Hazard, config_func=hazard_config
            ),
        ),
        ToolDef(
            "situation",
            gettext("Situation"),
            "gaphor-situation-symbolic",
            "s",
            new_item_factory(
                sysml_items.BlockItem,
                raaml.Situation,
                config_func=namespace_config,
            ),
        ),
        ToolDef(
            "control-structure",
            gettext("Control Structure"),
            "gaphor-control-structure-symbolic",
            "f",
            new_item_factory(
                sysml_items.BlockItem,
                raaml.ControlStructure,
                config_func=namespace_config,
            ),
        ),
        ToolDef(
            "controller",
            gettext("Controller"),
            "gaphor-controller-symbolic",
            "w",
            new_item_factory(
                sysml_items.PropertyItem,
                raaml.Controller,
                config_func=named_element_config,
            ),
        ),
        ToolDef(
            "actuator",
            gettext("Actuator"),
            "gaphor-actuator-symbolic",
            "q",
            new_item_factory(
                sysml_items.PropertyItem,
                raaml.Actuator,
                config_func=named_element_config,
            ),
        ),
        ToolDef(
            "controlled-process",
            gettext("Controlled Process"),
            "gaphor-controlled-process-symbolic",
            "<Shift>P",
            new_item_factory(
                sysml_items.PropertyItem,
                raaml.ControlledProcess,
                config_func=named_element_config,
            ),
        ),
        ToolDef(
            "abstract-operational-situation",
            gettext("Abstract Operational Situation"),
            "gaphor-abstract-operational-situation-symbolic",
            "<Shift>J",
            new_item_factory(
                diagramitems.OperationalSituationItem,
                raaml.AbstractOperationalSituation,
                config_func=abstract_operational_situation_config,
            ),
        ),
        ToolDef(
            "operational-situation",
            gettext("Operational Situation"),
            "gaphor-operational-situation-symbolic",
            "<Shift>O",
            new_item_factory(
                diagramitems.OperationalSituationItem,
                raaml.OperationalSituation,
                config_func=namespace_config,
            ),
        ),
        ToolDef(
            "unsafe-control-action",
            gettext("Unsafe Control Action"),
            "gaphor-unsafe-control-action-symbolic",
            "u",
            new_item_factory(
                diagramitems.UnsafeControlActionItem,
                raaml.UnsafeControlAction,
                config_func=namespace_config,
            ),
        ),
        ToolDef(
            "relevant-to",
            gettext("Relevant To"),
            "gaphor-relevant-to-symbolic",
            "r",
            new_item_factory(
                diagramitems.RelevantToItem,
            ),
        ),
        ToolDef(
            "control-action",
            gettext("Control Action"),
            "gaphor-control-action-symbolic",
            "<Shift>M",
            new_item_factory(
                diagramitems.ControlActionItem,
                raaml.ControlAction,
                config_func=namespace_config,
            ),
        ),
    ),
)



from gaphor import UML
from gaphor.core import transactional
from gaphor.diagram.propertypages import (
    PropertyPageBase,
    PropertyPages,
    combo_box_text_auto_complete,
    handler_blocking,
    new_resource_builder,
)

new_builder = new_resource_builder("gaphor.UML.profiles")


def _issubclass(c, b):
    try:
        return issubclass(c, b)
    except TypeError:
        return False


@PropertyPages.register(UML.Class)
class MetaclassPropertyPage(PropertyPageBase):
    """Adapter which shows a property page for a class view.

    Also handles metaclasses.
    """

    order = 10

    subject: UML.Class

    CLASSES = list(
        sorted(
            (c, c)
            for c in dir(UML)
            if _issubclass(getattr(UML, c), UML.Element) and c != "Stereotype"
        )
    )

    def __init__(self, subject):
        self.subject = subject
        self.watcher = subject.watcher()

    def construct(self):
        if not UML.recipes.is_metaclass(self.subject):
            return

        builder = new_builder(
            "metaclass-editor",
            signals={
                "metaclass-combo-changed": (self._on_name_changed,),
                "metaclass-combo-destroy": (self.watcher.unsubscribe_all,),
            },
        )

        combo = builder.get_object("metaclass-combo")
        combo_box_text_auto_complete(
            combo, self.CLASSES, self.subject and self.subject.name or ""
        )

        entry = combo.get_child()

        @handler_blocking(combo, "changed", self._on_name_changed)
        def handler(event):
            if event.element is self.subject and entry.get_text() != event.new_value:
                entry.set_text(event.new_value or "")

        self.watcher.watch("name", handler)

        return builder.get_object("metaclass-editor")

    @transactional
    def _on_name_changed(self, combo):
        self.subject.name = combo.get_active_text()



"""Undo management for Gaphor.

Undoing and redoing actions is managed through the UndoManager.

An undo action should be a callable object (called with no arguments).

An undo action should return a callable object that acts as redo function.
If None is returned the undo action is considered to be the redo action as well.

NOTE: it would be nice to use actions in conjunction with functools.partial.
"""

import logging
from typing import Callable, List

from gaphor.abc import ActionProvider, Service
from gaphor.action import action
from gaphor.core import event_handler
from gaphor.core.modeling.diagram import Diagram
from gaphor.core.modeling.element import Element, RepositoryProtocol
from gaphor.core.modeling.event import (
    AssociationAdded,
    AssociationDeleted,
    AssociationSet,
    AttributeUpdated,
    ElementCreated,
    ElementDeleted,
    ModelReady,
    RevertibleEvent,
)
from gaphor.core.modeling.presentation import Presentation
from gaphor.core.modeling.properties import association as association_property
from gaphor.diagram.copypaste import deserialize, serialize
from gaphor.event import (
    ActionEnabled,
    ServiceEvent,
    TransactionBegin,
    TransactionCommit,
    TransactionRollback,
)
from gaphor.transaction import Transaction, transactional

logger = logging.getLogger(__name__)


class ActionStack:
    """A transaction.

    Every action that is added between a begin_transaction() and a
    commit_transaction() call is recorded in a transaction, so it can be
    played back when a transaction is executed. This executing a
    transaction has the effect of performing the actions recorded, which
    will typically undo actions performed by the user.
    """

    def __init__(self):
        self._actions: List[Callable[[], None]] = []

    def add(self, action):
        self._actions.append(action)

    def can_execute(self):
        return bool(self._actions)

    @transactional
    def execute(self):
        self._actions.reverse()

        for act in self._actions:
            logger.debug(act.__doc__)
            act()


class UndoManagerStateChanged(ServiceEvent):
    """Event class used to send state changes on the Undo Manager."""


class NotInTransactionException(Exception):
    """Raised when changes occur outside a transaction."""


class UndoManager(Service, ActionProvider):
    """Simple transaction manager for Gaphor. This transaction manager supports
    nested transactions.

    The Undo manager sports an undo and a redo stack. Each stack
    contains a set of actions that can be executed, just by calling them
    (e.i action()) If something is returned by an action, that is
    considered the callable to be used to undo or redo the last
    performed action.
    """

    def __init__(self, event_manager, element_factory):
        self.event_manager = event_manager
        self.element_factory: RepositoryProtocol = element_factory
        self._undo_stack: List[ActionStack] = []
        self._redo_stack: List[ActionStack] = []
        self._stack_depth = 20
        self._current_transaction = None
        self._undoing = 0

        event_manager.subscribe(self.reset)
        event_manager.subscribe(self.begin_transaction)
        event_manager.subscribe(self.commit_transaction)
        event_manager.subscribe(self.rollback_transaction)
        self._register_undo_handlers()
        self._action_executed()

    def shutdown(self):
        self.event_manager.unsubscribe(self.reset)
        self.event_manager.unsubscribe(self.begin_transaction)
        self.event_manager.unsubscribe(self.commit_transaction)
        self.event_manager.unsubscribe(self.rollback_transaction)
        self._unregister_undo_handlers()

    def clear_undo_stack(self):
        self._undo_stack = []
        self._current_transaction = None

    def clear_redo_stack(self):
        del self._redo_stack[:]

    @event_handler(ModelReady)
    def reset(self, event=None):
        self.clear_redo_stack()
        self.clear_undo_stack()
        self._action_executed()

    @event_handler(TransactionBegin)
    def begin_transaction(self, event=None):
        """Add an action to the current transaction."""
        assert not self._current_transaction
        self._current_transaction = ActionStack()

    def add_undo_action(self, action, requires_transaction=True):
        """Add an action to undo."""
        if self._current_transaction:
            self._current_transaction.add(action)
            self._action_executed()
        elif requires_transaction:
            undo_stack = list(self._undo_stack)
            redo_stack = list(self._redo_stack)

            try:
                with Transaction(self.event_manager):
                    action()
            finally:
                # Restore stacks and act like nothing happened
                self._redo_stack = redo_stack
                self._undo_stack = undo_stack

            raise NotInTransactionException(
                f"Updating state outside of a transaction: {action.__doc__}."
            )

    @event_handler(TransactionCommit)
    def commit_transaction(self, event=None):
        assert self._current_transaction

        if self._current_transaction.can_execute():
            self.clear_redo_stack()
            self._undo_stack.append(self._current_transaction)
            while len(self._undo_stack) > self._stack_depth:
                del self._undo_stack[0]

        self._current_transaction = None

        self._action_executed()

    @event_handler(TransactionRollback)
    def rollback_transaction(self, event=None):
        """Roll back the transaction we're in."""
        assert self._current_transaction

        # Store stacks
        undo_stack = list(self._undo_stack)

        erroneous_tx = self._current_transaction
        self._current_transaction = None
        try:
            with Transaction(self.event_manager):
                try:
                    erroneous_tx.execute()
                except Exception:
                    logger.error("Could not rollback transaction", exc_info=True)
                    raise
        finally:
            # Discard all data collected in the rollback "transaction"
            self._undo_stack = undo_stack

        self._action_executed()

    def discard_transaction(self):

        self._current_transaction = None

        self._action_executed()

    @action(name="edit-undo", shortcut="<Primary>z")
    def undo_transaction(self):
        if not self._undo_stack:
            return

        if self._current_transaction:
            logger.warning("Trying to undo a transaction, while in a transaction")
            self.commit_transaction()
        transaction = self._undo_stack.pop()

        # Store stacks
        undo_stack = list(self._undo_stack)
        redo_stack = list(self._redo_stack)
        self._undo_stack = []

        try:
            self._undoing += 1
            with Transaction(self.event_manager):
                transaction.execute()
        finally:
            # Restore stacks and put latest tx on the redo stack
            self._redo_stack = redo_stack
            if self._undo_stack:
                self._redo_stack.extend(self._undo_stack)
            self._undo_stack = undo_stack
            self._undoing -= 1

        while len(self._redo_stack) > self._stack_depth:
            del self._redo_stack[0]

        self._action_executed()

    @action(
        name="edit-redo",
        shortcut="<Primary><Shift>z",
    )
    def redo_transaction(self):
        if not self._redo_stack:
            return

        transaction = self._redo_stack.pop()

        redo_stack = list(self._redo_stack)
        try:
            self._undoing += 1
            with Transaction(self.event_manager):
                transaction.execute()
        finally:
            self._redo_stack = redo_stack
            self._undoing -= 1

        self._action_executed()

    def in_transaction(self):
        """The undo manager is recording changes."""
        return self._current_transaction is not None

    def in_undo_transaction(self):
        """An undo or redo action is currently performed."""
        return bool(self._undoing)

    def can_undo(self):
        return bool(self._current_transaction or self._undo_stack)

    def can_redo(self):
        return bool(self._redo_stack)

    def _action_executed(self, event=None):
        self.event_manager.handle(ActionEnabled("win.edit-undo", self.can_undo()))
        self.event_manager.handle(ActionEnabled("win.edit-redo", self.can_redo()))
        self.event_manager.handle(UndoManagerStateChanged(self))

    def lookup(self, id: str) -> Element:
        if element := self.element_factory.lookup(id):
            return element
        else:
            raise ValueError(f"Element with id {id} not found in model")

    #
    # Undo Handlers
    #

    def _register_undo_handlers(self):
        self.event_manager.subscribe(self.undo_reversible_event)
        self.event_manager.subscribe(self.undo_create_element_event)
        self.event_manager.subscribe(self.undo_delete_element_event)
        self.event_manager.subscribe(self.undo_attribute_change_event)
        self.event_manager.subscribe(self.undo_association_set_event)
        self.event_manager.subscribe(self.undo_association_add_event)
        self.event_manager.subscribe(self.undo_association_delete_event)

    def _unregister_undo_handlers(self):
        self.event_manager.unsubscribe(self.undo_reversible_event)
        self.event_manager.unsubscribe(self.undo_create_element_event)
        self.event_manager.unsubscribe(self.undo_delete_element_event)
        self.event_manager.unsubscribe(self.undo_attribute_change_event)
        self.event_manager.unsubscribe(self.undo_association_set_event)
        self.event_manager.unsubscribe(self.undo_association_add_event)
        self.event_manager.unsubscribe(self.undo_association_delete_event)

    @event_handler(RevertibleEvent)
    def undo_reversible_event(self, event: RevertibleEvent):
        element_id = event.element.id

        def undo_reversible_event():
            element = self.lookup(element_id)
            event.revert(element)

        undo_reversible_event.__doc__ = (
            f"Reverse event {event.__class__.__name__} for element {event.element}."
        )

        self.add_undo_action(
            undo_reversible_event, requires_transaction=event.requires_transaction
        )

    @event_handler(ElementCreated)
    def undo_create_element_event(self, event: ElementCreated):
        element_id = event.element.id

        def undo_create_event():
            element = self.lookup(element_id)
            element.unlink()

        undo_create_event.__doc__ = f"Undo create element {event.element}."
        del event

        self.add_undo_action(undo_create_event)

    @event_handler(ElementDeleted)
    def undo_delete_element_event(self, event: ElementDeleted):
        element_type = type(event.element)
        element_id = event.element.id

        if isinstance(event.element, Presentation):
            diagram_id = event.diagram.id
            data = {}

            def save_func(name, value):
                data[name] = serialize(value)

            event.element.save(save_func)

            def undo_delete_event():
                # If diagram is not there, for some reason, recreate it.
                # It's probably removed in the same transaction.
                try:
                    diagram: Diagram = self.lookup(diagram_id)  # type: ignore[assignment]
                except ValueError:
                    diagram = self.element_factory.create_as(Diagram, diagram_id)

                element = diagram.create_as(element_type, element_id)
                for name, ser in data.items():
                    for value in deserialize(ser, lambda ref: None):
                        element.load(name, value)

        else:

            def undo_delete_event():
                self.element_factory.create_as(element_type, element_id)

        undo_delete_event.__doc__ = f"Recreate element {element_type} ({element_id})."
        del event

        self.add_undo_action(undo_delete_event)

    @event_handler(AttributeUpdated)
    def undo_attribute_change_event(self, event: AttributeUpdated):
        attribute = event.property
        element_id = event.element.id
        value = event.old_value

        def undo_attribute_change_event():
            element = self.lookup(element_id)
            attribute.set(element, value)

        undo_attribute_change_event.__doc__ = (
            f"Revert {event.element}.{attribute.name} to {value}."
        )
        del event

        self.add_undo_action(undo_attribute_change_event)

    @event_handler(AssociationSet)
    def undo_association_set_event(self, event: AssociationSet):
        association = event.property
        if type(association) is not association_property:
            return
        element_id = event.element.id
        value_id = event.old_value and event.old_value.id

        def undo_association_set_event():
            element = self.lookup(element_id)
            value = value_id and self.lookup(value_id)
            association.set(element, value, from_opposite=True)

        undo_association_set_event.__doc__ = (
            f"Revert {event.element}.{association.name} to {event.old_value}."
        )
        del event

        self.add_undo_action(undo_association_set_event)

    @event_handler(AssociationAdded)
    def undo_association_add_event(self, event: AssociationAdded):
        association = event.property
        if type(association) is not association_property:
            return
        element_id = event.element.id
        value_id = event.new_value.id

        def undo_association_add_event():
            element = self.lookup(element_id)
            value = self.lookup(value_id)
            association.delete(element, value, from_opposite=True)

        undo_association_add_event.__doc__ = (
            f"{event.element}.{association.name} delete {event.new_value}."
        )
        del event

        self.add_undo_action(undo_association_add_event)

    @event_handler(AssociationDeleted)
    def undo_association_delete_event(self, event: AssociationDeleted):
        association = event.property
        if type(association) is not association_property:
            return
        element_id = event.element.id
        value_id = event.old_value.id

        def undo_association_delete_event():
            element = self.lookup(element_id)
            value = self.lookup(value_id)
            association.set(element, value, from_opposite=True)

        undo_association_delete_event.__doc__ = (
            f"{event.element}.{association.name} add {event.old_value}."
        )
        del event

        self.add_undo_action(undo_association_delete_event)



# This file is generated by coder.py. DO NOT EDIT!
# isort: skip_file
# flake8: noqa F401,F811
# fmt: off

from __future__ import annotations

from gaphor.core.modeling.properties import (
    association,
    attribute as _attribute,
    derived,
    derivedunion,
    enumeration as _enumeration,
    redefine,
    relation_many,
    relation_one,
)


from gaphor.UML.uml import NamedElement
class AbstractRequirement(NamedElement):
    derived: relation_many[AbstractRequirement]
    derivedFrom: relation_many[AbstractRequirement]
    externalId: _attribute[str] = _attribute("externalId", str)
    master: relation_many[AbstractRequirement]
    refinedBy: relation_many[NamedElement]
    satisfiedBy: relation_many[NamedElement]
    text: _attribute[str] = _attribute("text", str)
    tracedTo: relation_many[NamedElement]
    verifiedBy: relation_many[NamedElement]


from gaphor.UML.uml import Class
class Requirement(AbstractRequirement, Class):
    pass


from gaphor.UML.uml import DirectedRelationship
class DirectedRelationshipPropertyPath(DirectedRelationship):
    sourceContext: relation_one[Classifier]
    sourcePropertyPath: relation_many[Property]
    targetContext: relation_one[Classifier]
    targetPropertyPath: relation_many[Property]


from gaphor.UML.uml import Dependency
class Trace(Dependency, DirectedRelationshipPropertyPath):
    pass


class Copy(Trace):
    pass


class Verify(Trace):
    pass


class DeriveReqt(Trace):
    pass


class Satisfy(Trace):
    pass


from gaphor.UML.uml import Behavior
class TestCase(Behavior):
    pass


class Block(Class):
    isEncapsulated: _attribute[int] = _attribute("isEncapsulated", int, default=False)


from gaphor.UML.uml import Property
class ConnectorProperty(Property):
    connector: relation_one[Connector]


class ParticipantProperty(Property):
    end_: relation_one[Property]


class DistributedProperty(Property):
    pass


from gaphor.UML.uml import DataType
class ValueType(DataType):
    quantityKind: relation_one[InstanceSpecification]
    unit: relation_one[InstanceSpecification]


from gaphor.UML.uml import InstanceSpecification
from gaphor.core.modeling.element import Element
class ElementPropertyPath(Element):
    propertyPath: relation_many[Property]


from gaphor.UML.uml import Connector
class BindingConnector(Connector):
    pass


from gaphor.UML.uml import ConnectorEnd
class NestedConnectorEnd(ConnectorEnd, ElementPropertyPath):
    pass


from gaphor.UML.uml import Classifier
class PropertySpecificType(Classifier):
    pass


class EndPathMultiplicity(Property):
    pass


class BoundReference(EndPathMultiplicity):
    bindingPath: relation_many[Property]
    boundend: relation_many[ConnectorEnd]


class AdjuntProperty(Property):
    principal: relation_one[Element]


from gaphor.UML.uml import Port
class ProxyPort(Port):
    pass


class FullPort(Port):
    pass


class FlowProperty(Property):
    direction = _enumeration("direction", ("in", "inout", "out"), "in")


class InterfaceBlock(Block):
    pass


from gaphor.UML.uml import InvocationAction
from gaphor.UML.uml import Trigger
from gaphor.UML.uml import AddStructuralFeatureValueAction
class InvocationOnNestedPortAction(ElementPropertyPath, InvocationAction):
    onNestedPort: relation_many[Port]


class TriggerOnNestedPort(ElementPropertyPath, Trigger):
    onNestedPort: relation_many[Port]


class AddFlowPropertyValueOnNestedPortAction(AddStructuralFeatureValueAction, ElementPropertyPath):
    pass


from gaphor.UML.uml import ChangeEvent
class ChangeSructuralFeatureEvent(ChangeEvent):
    structuralFeature: relation_one[StructuralFeature]


from gaphor.UML.uml import StructuralFeature
from gaphor.UML.uml import AcceptEventAction
class AcceptChangeStructuralFeatureEventAction(AcceptEventAction):
    pass


from gaphor.UML.uml import Feature
class DirectedFeature(Feature):
    featureDirection = _enumeration("featureDirection", ("provided", "providedRequired", "required"), "provided")


from gaphor.UML.uml import Generalization
class Conform(Generalization):
    pass


class View(Class):
    stakeholder: relation_many[Stakeholder]
    viewpoint: relation_one[Viewpoint]


class Viewpoint(Class):
    concernList: relation_many[Comment]
    language: _attribute[str] = _attribute("language", str)
    method: relation_many[Behavior]
    presentation: _attribute[str] = _attribute("presentation", str)
    purpose: _attribute[str] = _attribute("purpose", str)
    stakeholder: relation_many[Stakeholder]


class Stakeholder(Classifier):
    concernList: relation_many[Comment]


class Expose(Dependency):
    pass


from gaphor.core.modeling.coremodel import Comment
class Rationale(Comment):
    pass


class Problem(Comment):
    pass


class ElementGroup(Comment):
    member: relation_many[Element]
    name: _attribute[str] = _attribute("name", str)
    orderedMember: relation_many[Element]


class ConstraintBlock(Block):
    pass


from gaphor.UML.uml import Parameter
from gaphor.UML.uml import ActivityEdge
from gaphor.UML.uml import ParameterSet
class Optional(Parameter):
    pass


class Rate(ActivityEdge, Parameter):
    rate: relation_many[InstanceSpecification]


class Probability(ActivityEdge, ParameterSet):
    probability: _attribute[str] = _attribute("probability", str)


class Continuous(Rate):
    pass


class Discrete(Rate):
    pass


from gaphor.UML.uml import Operation
from gaphor.UML.uml import ObjectNode
class ControlOperator(Behavior):
    pass


class NoBuffer(ObjectNode):
    pass


class Overwrite(ObjectNode):
    pass


from gaphor.UML.uml import Abstraction
class Allocate(Abstraction, DirectedRelationshipPropertyPath):
    pass


from gaphor.UML.uml import ActivityPartition
class AllocateActivityPartition(ActivityPartition):
    pass


class Refine(Dependency, DirectedRelationshipPropertyPath):
    pass


class Tagged(Property):
    nonunique: _attribute[int] = _attribute("nonunique", int)
    ordered: _attribute[int] = _attribute("ordered", int)
    subsets: _attribute[str] = _attribute("subsets", str)


class ClassifierBehaviorProperty(Property):
    pass


from gaphor.UML.uml import InformationFlow
class ItemFlow(InformationFlow):
    itemProperty: relation_one[Property]



AbstractRequirement.derived = derivedunion("derived", AbstractRequirement)
AbstractRequirement.derivedFrom = derivedunion("derivedFrom", AbstractRequirement)
AbstractRequirement.master = derivedunion("master", AbstractRequirement)
AbstractRequirement.refinedBy = derivedunion("refinedBy", NamedElement)
AbstractRequirement.satisfiedBy = derivedunion("satisfiedBy", NamedElement)
AbstractRequirement.tracedTo = derivedunion("tracedTo", NamedElement)
AbstractRequirement.verifiedBy = derivedunion("verifiedBy", NamedElement)
DirectedRelationshipPropertyPath.targetContext = association("targetContext", Classifier, upper=1)
DirectedRelationshipPropertyPath.sourceContext = association("sourceContext", Classifier, upper=1)
DirectedRelationshipPropertyPath.sourcePropertyPath = association("sourcePropertyPath", Property)
DirectedRelationshipPropertyPath.targetPropertyPath = association("targetPropertyPath", Property)
Property.itemFlow = association("itemFlow", ItemFlow, upper=1, opposite="itemProperty")
from gaphor.UML.uml import Element
Element.owner.add(Property.itemFlow)  # type: ignore[attr-defined]
ConnectorProperty.connector = association("connector", Connector, upper=1, composite=True)
ParticipantProperty.end_ = association("end_", Property, upper=1, composite=True)
ValueType.unit = association("unit", InstanceSpecification, upper=1)
ValueType.quantityKind = association("quantityKind", InstanceSpecification, upper=1)
ElementPropertyPath.propertyPath = association("propertyPath", Property, lower=1)
BoundReference.boundend = association("boundend", ConnectorEnd, composite=True)
BoundReference.bindingPath = derivedunion("bindingPath", Property, lower=1)
AdjuntProperty.principal = association("principal", Element, upper=1)
InvocationOnNestedPortAction.onNestedPort = association("onNestedPort", Port, lower=1)
TriggerOnNestedPort.onNestedPort = association("onNestedPort", Port, lower=1)
ChangeSructuralFeatureEvent.structuralFeature = association("structuralFeature", StructuralFeature, upper=1)
View.stakeholder = derivedunion("stakeholder", Stakeholder)
View.viewpoint = derivedunion("viewpoint", Viewpoint, upper=1)
Viewpoint.concernList = association("concernList", Comment, composite=True)
Viewpoint.method = derivedunion("method", Behavior)
Viewpoint.stakeholder = association("stakeholder", Stakeholder, composite=True)
Stakeholder.concernList = association("concernList", Comment, composite=True)
ElementGroup.member = derivedunion("member", Element)
ElementGroup.orderedMember = association("orderedMember", Element, composite=True)
Rate.rate = association("rate", InstanceSpecification, composite=True)
ItemFlow.itemProperty = association("itemProperty", Property, upper=1, composite=True, opposite="itemFlow")
Element.ownedElement.add(ItemFlow.itemProperty)  # type: ignore[attr-defined]



from typing import Iterable, Optional, Set

from gaphas.item import Item
from gaphas.selection import Selection as _Selection


class Selection(_Selection):
    def __init__(self):
        super().__init__()
        self._dropzone_item: Optional[Item] = None
        self._grayed_out_items: Set[Item] = set()

    def clear(self):
        self._dropzone_item = None
        self._grayed_out_items.clear()
        super().clear()

    @property
    def dropzone_item(self) -> Optional[Item]:
        return self._dropzone_item

    @dropzone_item.setter
    def dropzone_item(self, item: Optional[Item]) -> None:
        if item is not self._dropzone_item:
            self._dropzone_item = item

    @property
    def grayed_out_items(self) -> Set[Item]:
        return self._grayed_out_items

    @grayed_out_items.setter
    def grayed_out_items(self, items: Iterable[Item]) -> None:
        self._grayed_out_items = set(items)



"""The element editor is a utility window used for editing elements."""

import logging
from typing import Optional

from gi.repository import GLib, Gtk, GtkSource

from gaphor.abc import ActionProvider
from gaphor.core import Transaction, action, event_handler
from gaphor.core.modeling import Presentation, StyleSheet
from gaphor.core.modeling.event import (
    AssociationUpdated,
    AttributeUpdated,
    ElementCreated,
    ModelReady,
)
from gaphor.diagram.propertypages import PropertyPages, new_resource_builder
from gaphor.ui.abc import UIComponent
from gaphor.ui.event import DiagramSelectionChanged

log = logging.getLogger(__name__)


new_builder = new_resource_builder("gaphor.ui", "elementeditor")


class DelayedFunction:
    def __init__(self, delay, function):
        self._delay = delay
        self._function = function
        self._source_id = 0

    def __call__(self, *args):
        if self._source_id:
            GLib.source_remove(self._source_id)

        def timeout_function():
            self._function(*args)
            self._source_id = 0

        self._source_id = GLib.timeout_add(self._delay, timeout_function)


class ElementEditor(UIComponent, ActionProvider):
    """The ElementEditor class is a utility window used to edit UML elements.

    It will display the properties of the currently selected element in
    the diagram.
    """

    def __init__(self, event_manager, element_factory, diagrams, properties):
        """Constructor.

        Build the action group for the element editor window. This will
        place a button for opening the window in the toolbar. The widget
        attribute is a PropertyEditor.
        """
        self.properties = properties
        self.editors = EditorStack(event_manager, diagrams, properties)
        self.preferences = PreferencesStack(event_manager, element_factory)

    def open(self):
        """Display the ElementEditor pane."""
        builder = new_builder("elementeditor", "style-sheet-buffer")

        self.revealer = builder.get_object("elementeditor")
        self.revealer.set_reveal_child(self.properties.get("show-editors", True))
        self.editor_stack = builder.get_object("editor-stack")

        self.editors.open(builder)
        self.preferences.open(builder)

        return self.revealer

    def close(self, widget=None):
        """Hide the element editor window and deactivate the toolbar button.

        Both the widget and event parameters default to None and are
        idempotent if set.
        """

        self.editors.close()
        self.preferences.close()

        return True

    @action(
        name="show-editors",
        shortcut="<Primary>e",
        state=lambda self: self.properties.get("show-editors", True),  # type: ignore[no-any-return]
    )
    def toggle_editor_visibility(self, active):
        self.revealer.set_reveal_child(active)
        self.properties.set("show-editors", active)

    @action(name="show-preferences", shortcut="<Primary>comma", state=False)
    def toggle_editor_preferences(self, active):
        if not self.revealer.get_child_revealed():
            if Gtk.get_major_version() == 3:
                self.revealer.get_action_group("win").lookup_action(
                    "show-editors"
                ).activate()
            else:
                self.revealer.activate_action("win.show-editors", None)

        self.editor_stack.set_visible_child_name("preferences" if active else "editors")

    @action(
        name="reset-tool-after-create",
        state=lambda self: self.properties.get("reset-tool-after-create", True),  # type: ignore[no-any-return]
    )
    def reset_tool_after_create(self, active):
        self.properties.set("reset-tool-after-create", active)

    @action(
        name="remove-unused-elements",
        state=lambda self: self.properties.get("remove-unused-elements", True),  # type: ignore[no-any-return]
    )
    def remove_unused_elements(self, active):
        self.properties.set("remove-unused-elements", active)


class EditorStack:
    def __init__(self, event_manager, diagrams, properties):
        self.event_manager = event_manager
        self.diagrams = diagrams
        self.properties = properties

        self.vbox: Optional[Gtk.Box] = None
        self._current_item = None
        self._expanded_pages = {}

    def open(self, builder):
        """Display the ElementEditor pane."""
        self.vbox = builder.get_object("editors")

        current_view = self.diagrams.get_current_view()
        self._selection_changed(
            focused_item=current_view and current_view.selection.focused_item
        )

        self.event_manager.subscribe(self._selection_changed)
        self.event_manager.subscribe(self._element_changed)

    def close(self):
        self.event_manager.unsubscribe(self._selection_changed)
        self.event_manager.unsubscribe(self._element_changed)

        self.vbox = None
        self._current_item = None

    def _get_adapters(self, item):
        """Return an ordered list of (order, name, adapter)."""
        adaptermap = {}
        if isinstance(item, Presentation) and item.subject:
            for adapter in PropertyPages(item.subject):
                adaptermap[(adapter.order, adapter.__class__.__name__)] = adapter
        for adapter in PropertyPages(item):
            adaptermap[(adapter.order, adapter.__class__.__name__)] = adapter

        return sorted(adaptermap.items())

    def create_pages(self, item):
        """Load all tabs that can operate on the given item."""
        assert self.vbox
        adapters = self._get_adapters(item)

        for (_, name), adapter in adapters:
            try:
                page = adapter.construct()
                if not page:
                    continue
                elif isinstance(page, Gtk.Expander):
                    page.set_expanded(self._expanded_pages.get(name, True))
                    page.connect_after("activate", self.on_expand, name)
                if Gtk.get_major_version() == 3:
                    self.vbox.pack_start(page, False, True, 0)
                else:
                    self.vbox.append(page)
            except Exception:
                log.error(
                    f"Could not construct property page for {name}", exc_info=True
                )

    def clear_pages(self):
        """Remove all tabs from the notebook."""
        assert self.vbox
        if Gtk.get_major_version() == 3:
            for page in self.vbox.get_children():
                page.destroy()
        else:
            while page := self.vbox.get_first_child():
                self.vbox.remove(page)

    def on_expand(self, widget, name):
        self._expanded_pages[name] = widget.get_expanded()

    @event_handler(DiagramSelectionChanged)
    def _selection_changed(self, event=None, focused_item=None):
        """Called when a diagram item receives focus.

        This reloads all tabs based on the current selection.
        """
        assert self.vbox
        item = event and event.focused_item or focused_item
        children = (
            self.vbox.get_children()
            if Gtk.get_major_version() == 3
            else self.vbox.get_first_child()
        )
        if item is self._current_item and children:
            return

        self._current_item = item
        self.clear_pages()

        if item:
            self.create_pages(item)
        else:
            self.show_no_item_selected()

    def show_no_item_selected(self):
        assert self.vbox
        builder = new_builder("no-item-selected")

        if Gtk.get_major_version() == 3:
            self.vbox.pack_start(
                child=builder.get_object("no-item-selected"),
                expand=False,
                fill=True,
                padding=0,
            )
        else:
            self.vbox.append(builder.get_object("no-item-selected"))

        tips = builder.get_object("tips")

        def on_show_tips_changed(checkbox, gparam):
            active = checkbox.get_active()
            tips.show() if active else tips.hide()
            self.properties.set("show-tips", active)

        show_tips = builder.get_object("show-tips")
        show_tips.connect("notify::active", on_show_tips_changed)
        show_tips.set_active(self.properties.get("show-tips", True))
        on_show_tips_changed(show_tips, None)

    @event_handler(AssociationUpdated)
    def _element_changed(self, event: AssociationUpdated):
        if event.property is Presentation.subject:  # type: ignore[misc] # noqa: F821
            element = event.element
            if element is self._current_item:
                self.clear_pages()
                self.create_pages(self._current_item)


class PreferencesStack:
    """Support code for the Preferences (cog) pane."""

    def __init__(self, event_manager, element_factory):
        self.event_manager = event_manager
        self.element_factory = element_factory
        self.lang_manager = GtkSource.LanguageManager.get_default()

        def tx_update_style_sheet(style_sheet, text):
            self._in_update = 1
            with Transaction(event_manager):
                style_sheet.styleSheet = text
            self._in_update = 0

        self._style_sheet_update = DelayedFunction(800, tx_update_style_sheet)
        self._in_update = 0

    def open(self, builder):
        self.style_sheet_buffer = builder.get_object("style-sheet-buffer")
        self.style_sheet_view = builder.get_object("style-sheet-view")

        self.style_sheet_buffer.set_language(self.lang_manager.get_language("css"))

        self.event_manager.subscribe(self._model_ready)
        self.event_manager.subscribe(self._style_sheet_created)
        self.event_manager.subscribe(self._style_sheet_changed)
        self.style_sheet_buffer.connect("changed", self.on_style_sheet_changed)
        self.update_text()

    def close(self):
        self.event_manager.unsubscribe(self._model_ready)
        self.event_manager.unsubscribe(self._style_sheet_changed)
        self.event_manager.unsubscribe(self._style_sheet_created)

    @property
    def style_sheet(self):
        return next(self.element_factory.select(StyleSheet), None)

    def on_style_sheet_changed(self, buffer):
        style_sheet = self.style_sheet
        if style_sheet and not self._in_update:
            text = buffer.get_text(
                buffer.get_start_iter(), buffer.get_end_iter(), False
            )

            self._style_sheet_update(style_sheet, text)

    def update_text(self):
        style_sheet = self.style_sheet
        if not style_sheet or self._in_update:
            return

        self._in_update = 1
        self.style_sheet_buffer.set_text(style_sheet.styleSheet or "")
        self._in_update = 0

    @event_handler(ModelReady)
    def _model_ready(self, event):
        self.update_text()

    @event_handler(ElementCreated)
    def _style_sheet_created(self, event):
        if isinstance(event.element, StyleSheet):
            self.update_text()

    @event_handler(AttributeUpdated)
    def _style_sheet_changed(self, event):
        if event.property is StyleSheet.styleSheet:
            self.update_text()



from gaphor.core import transactional
from gaphor.diagram.propertypages import (
    ComboModel,
    PropertyPageBase,
    PropertyPages,
    new_resource_builder,
)
from gaphor.UML.interactions.interactionsconnect import get_lifeline
from gaphor.UML.interactions.message import MessageItem

new_builder = new_resource_builder("gaphor.UML.interactions")


@PropertyPages.register(MessageItem)
class MessagePropertyPage(PropertyPageBase):
    """Property page for editing message items.

    When message is on communication diagram, then additional messages
    can be added. On sequence diagram sort of message can be changed.
    """

    order = 15

    MESSAGE_SORT = [
        ("Call", "synchCall"),
        ("Asynchronous", "asynchCall"),
        ("Signal", "asynchSignal"),
        ("Create", "createMessage"),
        ("Delete", "deleteMessage"),
        ("Reply", "reply"),
    ]

    def __init__(self, item):
        self.item = item

    def construct(self):
        item = self.item
        subject = item.subject

        if not subject or item.is_communication():
            return

        builder = new_builder(
            "message-editor",
            signals={"message-combo-changed": (self._on_message_sort_change,)},
        )

        sort_data = self.MESSAGE_SORT
        lifeline = get_lifeline(item, item.tail)

        # disallow connecting two delete messages to a lifeline
        if (
            lifeline
            and lifeline.is_destroyed
            and subject.messageSort != "deleteMessage"
        ):
            sort_data = sort_data.copy()
            assert sort_data[4][1] == "deleteMessage"
            del sort_data[4]

        self.model = ComboModel(sort_data)
        combo = builder.get_object("message-combo")
        combo.set_model(self.model)

        index = self.model.get_index(subject.messageSort)
        combo.set_active(index)

        return builder.get_object("message-editor")

    @transactional
    def _on_message_sort_change(self, combo):
        """Update message item's message sort information."""

        ms = self.model.get_value(combo.get_active())

        item = self.item
        subject = item.subject
        lifeline = get_lifeline(item, item.tail)

        # allow only one delete message to connect to lifeline's lifetime
        # destroyed status can be changed only by delete message itself
        if lifeline and (
            subject.messageSort == "deleteMessage" or not lifeline.is_destroyed
        ):
            is_destroyed = ms == "deleteMessage"
            lifeline.is_destroyed = is_destroyed
            lifeline.request_update()

        subject.messageSort = ms
        item.request_update()



"""Magnet tool pulls or pushes the right or bottom side depending on mouse
movement."""


import contextlib

from gaphas.canvas import ancestors
from gaphas.move import Move
from gaphas.view import GtkView
from gi.repository import Gtk

from gaphor.diagram.event import ToolCompleted


def magnet_tool(view: GtkView, event_manager) -> Gtk.GestureDrag:
    """Handle item movement and movement of handles."""
    gesture = (
        Gtk.GestureDrag.new(view)
        if Gtk.get_major_version() == 3
        else Gtk.GestureDrag.new()
    )
    drag_state = DragState(event_manager)
    gesture.connect("drag-begin", on_drag_begin, drag_state)
    gesture.connect("drag-update", on_drag_update, drag_state)
    gesture.connect("drag-end", on_drag_end, drag_state)
    return gesture


class DragState:
    def __init__(self, event_manager):
        self.event_manager = event_manager
        self.reset()

    def reset(self):
        self.direction = 0
        self.moving_items = set()


def on_drag_begin(gesture, start_x, start_y, drag_state):
    drag_state.reset()
    gesture.set_state(Gtk.EventSequenceState.CLAIMED)


def on_drag_update(gesture, offset_x, offset_y, drag_state):
    _, sx, sy = gesture.get_start_point()
    view = gesture.get_widget()

    if not drag_state.direction:
        if abs(offset_x) > 10 and abs(offset_x) > abs(offset_y):
            drag_state.direction = 1
            drag_state.moving_items = set(
                moving_items(
                    view,
                    set(view.get_items_in_rectangle((sx, -100000, 10000000, 1000000))),
                )
            )
            for moving in drag_state.moving_items:
                moving.start_move((sx, 0))
        elif abs(offset_y) > 10 and abs(offset_y) > abs(offset_x):
            drag_state.direction = 2
            drag_state.moving_items = set(
                moving_items(
                    view,
                    set(view.get_items_in_rectangle((-100000, sy, 10000000, 1000000))),
                )
            )
            for moving in drag_state.moving_items:
                moving.start_move((0, sy))
        else:
            return

    if drag_state.direction == 1:
        x = sx + offset_x
        y = 0
    elif drag_state.direction == 2:
        x = 0
        y = sy + offset_y

    for moving in drag_state.moving_items:
        moving.move((x, y))

    view.magnet = (drag_state.direction, x, y)
    view.update_back_buffer()


def moving_items(view, selected_items):
    for item in selected_items:
        # Do not move subitems of selected items
        if not set(ancestors(view.model, item)).intersection(selected_items):
            yield Move(item, view)


def on_drag_end(gesture, offset_x, offset_y, drag_state):
    view = gesture.get_widget()
    _, x, y = gesture.get_start_point()
    for moving in drag_state.moving_items:
        moving.stop_move((x + offset_x, y + offset_y))
    drag_state.reset()
    view.update_back_buffer()
    drag_state.event_manager.handle(ToolCompleted())
    with contextlib.suppress(AttributeError):
        del view.magnet


class MagnetPainter:
    def __init__(self, view: GtkView):
        self.view = view

    def paint(self, _items, cr):
        try:
            magnet = self.view.magnet
        except AttributeError:
            return

        view = self.view
        allocation = view.get_allocation()
        w, h = allocation.width, allocation.height
        direction, x, y = magnet

        cr.save()
        try:
            cr.identity_matrix()
            cr.set_line_width(6)
            cr.set_source_rgba(0.0, 0.3, 0.3, 0.3)
            if direction == 1:
                cr.move_to(x, 0)
                cr.line_to(x, h)
            elif direction == 2:
                cr.move_to(0, y)
                cr.line_to(w, y)
            cr.stroke()
        finally:
            cr.restore()
